# -*- coding: utf-8 -*-
#
# NOTE: This file was auto-generated with MetaTools/buildUCD.py.
# Source: https://unicode.org/Public/UNIDATA/Blocks.txt
# License: http://unicode.org/copyright.html#License
#
# Blocks-13.0.0.txt
# Date: 2019-07-10, 19:06:00 GMT [KW]
# © 2019 Unicode®, Inc.
# For terms of use, see http://www.unicode.org/terms_of_use.html
#
# Unicode Character Database
# For documentation, see http://www.unicode.org/reports/tr44/
#
# Format:
# Start Code..End Code; Block Name


RANGES = [
    0x0000,  # .. 0x007F ; Basic Latin
    0x0080,  # .. 0x00FF ; Latin-1 Supplement
    0x0100,  # .. 0x017F ; Latin Extended-A
    0x0180,  # .. 0x024F ; Latin Extended-B
    0x0250,  # .. 0x02AF ; IPA Extensions
    0x02B0,  # .. 0x02FF ; Spacing Modifier Letters
    0x0300,  # .. 0x036F ; Combining Diacritical Marks
    0x0370,  # .. 0x03FF ; Greek and Coptic
    0x0400,  # .. 0x04FF ; Cyrillic
    0x0500,  # .. 0x052F ; Cyrillic Supplement
    0x0530,  # .. 0x058F ; Armenian
    0x0590,  # .. 0x05FF ; Hebrew
    0x0600,  # .. 0x06FF ; Arabic
    0x0700,  # .. 0x074F ; Syriac
    0x0750,  # .. 0x077F ; Arabic Supplement
    0x0780,  # .. 0x07BF ; Thaana
    0x07C0,  # .. 0x07FF ; NKo
    0x0800,  # .. 0x083F ; Samaritan
    0x0840,  # .. 0x085F ; Mandaic
    0x0860,  # .. 0x086F ; Syriac Supplement
    0x0870,  # .. 0x089F ; No_Block
    0x08A0,  # .. 0x08FF ; Arabic Extended-A
    0x0900,  # .. 0x097F ; Devanagari
    0x0980,  # .. 0x09FF ; Bengali
    0x0A00,  # .. 0x0A7F ; Gurmukhi
    0x0A80,  # .. 0x0AFF ; Gujarati
    0x0B00,  # .. 0x0B7F ; Oriya
    0x0B80,  # .. 0x0BFF ; Tamil
    0x0C00,  # .. 0x0C7F ; Telugu
    0x0C80,  # .. 0x0CFF ; Kannada
    0x0D00,  # .. 0x0D7F ; Malayalam
    0x0D80,  # .. 0x0DFF ; Sinhala
    0x0E00,  # .. 0x0E7F ; Thai
    0x0E80,  # .. 0x0EFF ; Lao
    0x0F00,  # .. 0x0FFF ; Tibetan
    0x1000,  # .. 0x109F ; Myanmar
    0x10A0,  # .. 0x10FF ; Georgian
    0x1100,  # .. 0x11FF ; Hangul Jamo
    0x1200,  # .. 0x137F ; Ethiopic
    0x1380,  # .. 0x139F ; Ethiopic Supplement
    0x13A0,  # .. 0x13FF ; Cherokee
    0x1400,  # .. 0x167F ; Unified Canadian Aboriginal Syllabics
    0x1680,  # .. 0x169F ; Ogham
    0x16A0,  # .. 0x16FF ; Runic
    0x1700,  # .. 0x171F ; Tagalog
    0x1720,  # .. 0x173F ; Hanunoo
    0x1740,  # .. 0x175F ; Buhid
    0x1760,  # .. 0x177F ; Tagbanwa
    0x1780,  # .. 0x17FF ; Khmer
    0x1800,  # .. 0x18AF ; Mongolian
    0x18B0,  # .. 0x18FF ; Unified Canadian Aboriginal Syllabics Extended
    0x1900,  # .. 0x194F ; Limbu
    0x1950,  # .. 0x197F ; Tai Le
    0x1980,  # .. 0x19DF ; New Tai Lue
    0x19E0,  # .. 0x19FF ; Khmer Symbols
    0x1A00,  # .. 0x1A1F ; Buginese
    0x1A20,  # .. 0x1AAF ; Tai Tham
    0x1AB0,  # .. 0x1AFF ; Combining Diacritical Marks Extended
    0x1B00,  # .. 0x1B7F ; Balinese
    0x1B80,  # .. 0x1BBF ; Sundanese
    0x1BC0,  # .. 0x1BFF ; Batak
    0x1C00,  # .. 0x1C4F ; Lepcha
    0x1C50,  # .. 0x1C7F ; Ol Chiki
    0x1C80,  # .. 0x1C8F ; Cyrillic Extended-C
    0x1C90,  # .. 0x1CBF ; Georgian Extended
    0x1CC0,  # .. 0x1CCF ; Sundanese Supplement
    0x1CD0,  # .. 0x1CFF ; Vedic Extensions
    0x1D00,  # .. 0x1D7F ; Phonetic Extensions
    0x1D80,  # .. 0x1DBF ; Phonetic Extensions Supplement
    0x1DC0,  # .. 0x1DFF ; Combining Diacritical Marks Supplement
    0x1E00,  # .. 0x1EFF ; Latin Extended Additional
    0x1F00,  # .. 0x1FFF ; Greek Extended
    0x2000,  # .. 0x206F ; General Punctuation
    0x2070,  # .. 0x209F ; Superscripts and Subscripts
    0x20A0,  # .. 0x20CF ; Currency Symbols
    0x20D0,  # .. 0x20FF ; Combining Diacritical Marks for Symbols
    0x2100,  # .. 0x214F ; Letterlike Symbols
    0x2150,  # .. 0x218F ; Number Forms
    0x2190,  # .. 0x21FF ; Arrows
    0x2200,  # .. 0x22FF ; Mathematical Operators
    0x2300,  # .. 0x23FF ; Miscellaneous Technical
    0x2400,  # .. 0x243F ; Control Pictures
    0x2440,  # .. 0x245F ; Optical Character Recognition
    0x2460,  # .. 0x24FF ; Enclosed Alphanumerics
    0x2500,  # .. 0x257F ; Box Drawing
    0x2580,  # .. 0x259F ; Block Elements
    0x25A0,  # .. 0x25FF ; Geometric Shapes
    0x2600,  # .. 0x26FF ; Miscellaneous Symbols
    0x2700,  # .. 0x27BF ; Dingbats
    0x27C0,  # .. 0x27EF ; Miscellaneous Mathematical Symbols-A
    0x27F0,  # .. 0x27FF ; Supplemental Arrows-A
    0x2800,  # .. 0x28FF ; Braille Patterns
    0x2900,  # .. 0x297F ; Supplemental Arrows-B
    0x2980,  # .. 0x29FF ; Miscellaneous Mathematical Symbols-B
    0x2A00,  # .. 0x2AFF ; Supplemental Mathematical Operators
    0x2B00,  # .. 0x2BFF ; Miscellaneous Symbols and Arrows
    0x2C00,  # .. 0x2C5F ; Glagolitic
    0x2C60,  # .. 0x2C7F ; Latin Extended-C
    0x2C80,  # .. 0x2CFF ; Coptic
    0x2D00,  # .. 0x2D2F ; Georgian Supplement
    0x2D30,  # .. 0x2D7F ; Tifinagh
    0x2D80,  # .. 0x2DDF ; Ethiopic Extended
    0x2DE0,  # .. 0x2DFF ; Cyrillic Extended-A
    0x2E00,  # .. 0x2E7F ; Supplemental Punctuation
    0x2E80,  # .. 0x2EFF ; CJK Radicals Supplement
    0x2F00,  # .. 0x2FDF ; Kangxi Radicals
    0x2FE0,  # .. 0x2FEF ; No_Block
    0x2FF0,  # .. 0x2FFF ; Ideographic Description Characters
    0x3000,  # .. 0x303F ; CJK Symbols and Punctuation
    0x3040,  # .. 0x309F ; Hiragana
    0x30A0,  # .. 0x30FF ; Katakana
    0x3100,  # .. 0x312F ; Bopomofo
    0x3130,  # .. 0x318F ; Hangul Compatibility Jamo
    0x3190,  # .. 0x319F ; Kanbun
    0x31A0,  # .. 0x31BF ; Bopomofo Extended
    0x31C0,  # .. 0x31EF ; CJK Strokes
    0x31F0,  # .. 0x31FF ; Katakana Phonetic Extensions
    0x3200,  # .. 0x32FF ; Enclosed CJK Letters and Months
    0x3300,  # .. 0x33FF ; CJK Compatibility
    0x3400,  # .. 0x4DBF ; CJK Unified Ideographs Extension A
    0x4DC0,  # .. 0x4DFF ; Yijing Hexagram Symbols
    0x4E00,  # .. 0x9FFF ; CJK Unified Ideographs
    0xA000,  # .. 0xA48F ; Yi Syllables
    0xA490,  # .. 0xA4CF ; Yi Radicals
    0xA4D0,  # .. 0xA4FF ; Lisu
    0xA500,  # .. 0xA63F ; Vai
    0xA640,  # .. 0xA69F ; Cyrillic Extended-B
    0xA6A0,  # .. 0xA6FF ; Bamum
    0xA700,  # .. 0xA71F ; Modifier Tone Letters
    0xA720,  # .. 0xA7FF ; Latin Extended-D
    0xA800,  # .. 0xA82F ; Syloti Nagri
    0xA830,  # .. 0xA83F ; Common Indic Number Forms
    0xA840,  # .. 0xA87F ; Phags-pa
    0xA880,  # .. 0xA8DF ; Saurashtra
    0xA8E0,  # .. 0xA8FF ; Devanagari Extended
    0xA900,  # .. 0xA92F ; Kayah Li
    0xA930,  # .. 0xA95F ; Rejang
    0xA960,  # .. 0xA97F ; Hangul Jamo Extended-A
    0xA980,  # .. 0xA9DF ; Javanese
    0xA9E0,  # .. 0xA9FF ; Myanmar Extended-B
    0xAA00,  # .. 0xAA5F ; Cham
    0xAA60,  # .. 0xAA7F ; Myanmar Extended-A
    0xAA80,  # .. 0xAADF ; Tai Viet
    0xAAE0,  # .. 0xAAFF ; Meetei Mayek Extensions
    0xAB00,  # .. 0xAB2F ; Ethiopic Extended-A
    0xAB30,  # .. 0xAB6F ; Latin Extended-E
    0xAB70,  # .. 0xABBF ; Cherokee Supplement
    0xABC0,  # .. 0xABFF ; Meetei Mayek
    0xAC00,  # .. 0xD7AF ; Hangul Syllables
    0xD7B0,  # .. 0xD7FF ; Hangul Jamo Extended-B
    0xD800,  # .. 0xDB7F ; High Surrogates
    0xDB80,  # .. 0xDBFF ; High Private Use Surrogates
    0xDC00,  # .. 0xDFFF ; Low Surrogates
    0xE000,  # .. 0xF8FF ; Private Use Area
    0xF900,  # .. 0xFAFF ; CJK Compatibility Ideographs
    0xFB00,  # .. 0xFB4F ; Alphabetic Presentation Forms
    0xFB50,  # .. 0xFDFF ; Arabic Presentation Forms-A
    0xFE00,  # .. 0xFE0F ; Variation Selectors
    0xFE10,  # .. 0xFE1F ; Vertical Forms
    0xFE20,  # .. 0xFE2F ; Combining Half Marks
    0xFE30,  # .. 0xFE4F ; CJK Compatibility Forms
    0xFE50,  # .. 0xFE6F ; Small Form Variants
    0xFE70,  # .. 0xFEFF ; Arabic Presentation Forms-B
    0xFF00,  # .. 0xFFEF ; Halfwidth and Fullwidth Forms
    0xFFF0,  # .. 0xFFFF ; Specials
    0x10000,  # .. 0x1007F ; Linear B Syllabary
    0x10080,  # .. 0x100FF ; Linear B Ideograms
    0x10100,  # .. 0x1013F ; Aegean Numbers
    0x10140,  # .. 0x1018F ; Ancient Greek Numbers
    0x10190,  # .. 0x101CF ; Ancient Symbols
    0x101D0,  # .. 0x101FF ; Phaistos Disc
    0x10200,  # .. 0x1027F ; No_Block
    0x10280,  # .. 0x1029F ; Lycian
    0x102A0,  # .. 0x102DF ; Carian
    0x102E0,  # .. 0x102FF ; Coptic Epact Numbers
    0x10300,  # .. 0x1032F ; Old Italic
    0x10330,  # .. 0x1034F ; Gothic
    0x10350,  # .. 0x1037F ; Old Permic
    0x10380,  # .. 0x1039F ; Ugaritic
    0x103A0,  # .. 0x103DF ; Old Persian
    0x103E0,  # .. 0x103FF ; No_Block
    0x10400,  # .. 0x1044F ; Deseret
    0x10450,  # .. 0x1047F ; Shavian
    0x10480,  # .. 0x104AF ; Osmanya
    0x104B0,  # .. 0x104FF ; Osage
    0x10500,  # .. 0x1052F ; Elbasan
    0x10530,  # .. 0x1056F ; Caucasian Albanian
    0x10570,  # .. 0x105FF ; No_Block
    0x10600,  # .. 0x1077F ; Linear A
    0x10780,  # .. 0x107FF ; No_Block
    0x10800,  # .. 0x1083F ; Cypriot Syllabary
    0x10840,  # .. 0x1085F ; Imperial Aramaic
    0x10860,  # .. 0x1087F ; Palmyrene
    0x10880,  # .. 0x108AF ; Nabataean
    0x108B0,  # .. 0x108DF ; No_Block
    0x108E0,  # .. 0x108FF ; Hatran
    0x10900,  # .. 0x1091F ; Phoenician
    0x10920,  # .. 0x1093F ; Lydian
    0x10940,  # .. 0x1097F ; No_Block
    0x10980,  # .. 0x1099F ; Meroitic Hieroglyphs
    0x109A0,  # .. 0x109FF ; Meroitic Cursive
    0x10A00,  # .. 0x10A5F ; Kharoshthi
    0x10A60,  # .. 0x10A7F ; Old South Arabian
    0x10A80,  # .. 0x10A9F ; Old North Arabian
    0x10AA0,  # .. 0x10ABF ; No_Block
    0x10AC0,  # .. 0x10AFF ; Manichaean
    0x10B00,  # .. 0x10B3F ; Avestan
    0x10B40,  # .. 0x10B5F ; Inscriptional Parthian
    0x10B60,  # .. 0x10B7F ; Inscriptional Pahlavi
    0x10B80,  # .. 0x10BAF ; Psalter Pahlavi
    0x10BB0,  # .. 0x10BFF ; No_Block
    0x10C00,  # .. 0x10C4F ; Old Turkic
    0x10C50,  # .. 0x10C7F ; No_Block
    0x10C80,  # .. 0x10CFF ; Old Hungarian
    0x10D00,  # .. 0x10D3F ; Hanifi Rohingya
    0x10D40,  # .. 0x10E5F ; No_Block
    0x10E60,  # .. 0x10E7F ; Rumi Numeral Symbols
    0x10E80,  # .. 0x10EBF ; Yezidi
    0x10EC0,  # .. 0x10EFF ; No_Block
    0x10F00,  # .. 0x10F2F ; Old Sogdian
    0x10F30,  # .. 0x10F6F ; Sogdian
    0x10F70,  # .. 0x10FAF ; No_Block
    0x10FB0,  # .. 0x10FDF ; Chorasmian
    0x10FE0,  # .. 0x10FFF ; Elymaic
    0x11000,  # .. 0x1107F ; Brahmi
    0x11080,  # .. 0x110CF ; Kaithi
    0x110D0,  # .. 0x110FF ; Sora Sompeng
    0x11100,  # .. 0x1114F ; Chakma
    0x11150,  # .. 0x1117F ; Mahajani
    0x11180,  # .. 0x111DF ; Sharada
    0x111E0,  # .. 0x111FF ; Sinhala Archaic Numbers
    0x11200,  # .. 0x1124F ; Khojki
    0x11250,  # .. 0x1127F ; No_Block
    0x11280,  # .. 0x112AF ; Multani
    0x112B0,  # .. 0x112FF ; Khudawadi
    0x11300,  # .. 0x1137F ; Grantha
    0x11380,  # .. 0x113FF ; No_Block
    0x11400,  # .. 0x1147F ; Newa
    0x11480,  # .. 0x114DF ; Tirhuta
    0x114E0,  # .. 0x1157F ; No_Block
    0x11580,  # .. 0x115FF ; Siddham
    0x11600,  # .. 0x1165F ; Modi
    0x11660,  # .. 0x1167F ; Mongolian Supplement
    0x11680,  # .. 0x116CF ; Takri
    0x116D0,  # .. 0x116FF ; No_Block
    0x11700,  # .. 0x1173F ; Ahom
    0x11740,  # .. 0x117FF ; No_Block
    0x11800,  # .. 0x1184F ; Dogra
    0x11850,  # .. 0x1189F ; No_Block
    0x118A0,  # .. 0x118FF ; Warang Citi
    0x11900,  # .. 0x1195F ; Dives Akuru
    0x11960,  # .. 0x1199F ; No_Block
    0x119A0,  # .. 0x119FF ; Nandinagari
    0x11A00,  # .. 0x11A4F ; Zanabazar Square
    0x11A50,  # .. 0x11AAF ; Soyombo
    0x11AB0,  # .. 0x11ABF ; No_Block
    0x11AC0,  # .. 0x11AFF ; Pau Cin Hau
    0x11B00,  # .. 0x11BFF ; No_Block
    0x11C00,  # .. 0x11C6F ; Bhaiksuki
    0x11C70,  # .. 0x11CBF ; Marchen
    0x11CC0,  # .. 0x11CFF ; No_Block
    0x11D00,  # .. 0x11D5F ; Masaram Gondi
    0x11D60,  # .. 0x11DAF ; Gunjala Gondi
    0x11DB0,  # .. 0x11EDF ; No_Block
    0x11EE0,  # .. 0x11EFF ; Makasar
    0x11F00,  # .. 0x11FAF ; No_Block
    0x11FB0,  # .. 0x11FBF ; Lisu Supplement
    0x11FC0,  # .. 0x11FFF ; Tamil Supplement
    0x12000,  # .. 0x123FF ; Cuneiform
    0x12400,  # .. 0x1247F ; Cuneiform Numbers and Punctuation
    0x12480,  # .. 0x1254F ; Early Dynastic Cuneiform
    0x12550,  # .. 0x12FFF ; No_Block
    0x13000,  # .. 0x1342F ; Egyptian Hieroglyphs
    0x13430,  # .. 0x1343F ; Egyptian Hieroglyph Format Controls
    0x13440,  # .. 0x143FF ; No_Block
    0x14400,  # .. 0x1467F ; Anatolian Hieroglyphs
    0x14680,  # .. 0x167FF ; No_Block
    0x16800,  # .. 0x16A3F ; Bamum Supplement
    0x16A40,  # .. 0x16A6F ; Mro
    0x16A70,  # .. 0x16ACF ; No_Block
    0x16AD0,  # .. 0x16AFF ; Bassa Vah
    0x16B00,  # .. 0x16B8F ; Pahawh Hmong
    0x16B90,  # .. 0x16E3F ; No_Block
    0x16E40,  # .. 0x16E9F ; Medefaidrin
    0x16EA0,  # .. 0x16EFF ; No_Block
    0x16F00,  # .. 0x16F9F ; Miao
    0x16FA0,  # .. 0x16FDF ; No_Block
    0x16FE0,  # .. 0x16FFF ; Ideographic Symbols and Punctuation
    0x17000,  # .. 0x187FF ; Tangut
    0x18800,  # .. 0x18AFF ; Tangut Components
    0x18B00,  # .. 0x18CFF ; Khitan Small Script
    0x18D00,  # .. 0x18D8F ; Tangut Supplement
    0x18D90,  # .. 0x1AFFF ; No_Block
    0x1B000,  # .. 0x1B0FF ; Kana Supplement
    0x1B100,  # .. 0x1B12F ; Kana Extended-A
    0x1B130,  # .. 0x1B16F ; Small Kana Extension
    0x1B170,  # .. 0x1B2FF ; Nushu
    0x1B300,  # .. 0x1BBFF ; No_Block
    0x1BC00,  # .. 0x1BC9F ; Duployan
    0x1BCA0,  # .. 0x1BCAF ; Shorthand Format Controls
    0x1BCB0,  # .. 0x1CFFF ; No_Block
    0x1D000,  # .. 0x1D0FF ; Byzantine Musical Symbols
    0x1D100,  # .. 0x1D1FF ; Musical Symbols
    0x1D200,  # .. 0x1D24F ; Ancient Greek Musical Notation
    0x1D250,  # .. 0x1D2DF ; No_Block
    0x1D2E0,  # .. 0x1D2FF ; Mayan Numerals
    0x1D300,  # .. 0x1D35F ; Tai Xuan Jing Symbols
    0x1D360,  # .. 0x1D37F ; Counting Rod Numerals
    0x1D380,  # .. 0x1D3FF ; No_Block
    0x1D400,  # .. 0x1D7FF ; Mathematical Alphanumeric Symbols
    0x1D800,  # .. 0x1DAAF ; Sutton SignWriting
    0x1DAB0,  # .. 0x1DFFF ; No_Block
    0x1E000,  # .. 0x1E02F ; Glagolitic Supplement
    0x1E030,  # .. 0x1E0FF ; No_Block
    0x1E100,  # .. 0x1E14F ; Nyiakeng Puachue Hmong
    0x1E150,  # .. 0x1E2BF ; No_Block
    0x1E2C0,  # .. 0x1E2FF ; Wancho
    0x1E300,  # .. 0x1E7FF ; No_Block
    0x1E800,  # .. 0x1E8DF ; Mende Kikakui
    0x1E8E0,  # .. 0x1E8FF ; No_Block
    0x1E900,  # .. 0x1E95F ; Adlam
    0x1E960,  # .. 0x1EC6F ; No_Block
    0x1EC70,  # .. 0x1ECBF ; Indic Siyaq Numbers
    0x1ECC0,  # .. 0x1ECFF ; No_Block
    0x1ED00,  # .. 0x1ED4F ; Ottoman Siyaq Numbers
    0x1ED50,  # .. 0x1EDFF ; No_Block
    0x1EE00,  # .. 0x1EEFF ; Arabic Mathematical Alphabetic Symbols
    0x1EF00,  # .. 0x1EFFF ; No_Block
    0x1F000,  # .. 0x1F02F ; Mahjong Tiles
    0x1F030,  # .. 0x1F09F ; Domino Tiles
    0x1F0A0,  # .. 0x1F0FF ; Playing Cards
    0x1F100,  # .. 0x1F1FF ; Enclosed Alphanumeric Supplement
    0x1F200,  # .. 0x1F2FF ; Enclosed Ideographic Supplement
    0x1F300,  # .. 0x1F5FF ; Miscellaneous Symbols and Pictographs
    0x1F600,  # .. 0x1F64F ; Emoticons
    0x1F650,  # .. 0x1F67F ; Ornamental Dingbats
    0x1F680,  # .. 0x1F6FF ; Transport and Map Symbols
    0x1F700,  # .. 0x1F77F ; Alchemical Symbols
    0x1F780,  # .. 0x1F7FF ; Geometric Shapes Extended
    0x1F800,  # .. 0x1F8FF ; Supplemental Arrows-C
    0x1F900,  # .. 0x1F9FF ; Supplemental Symbols and Pictographs
    0x1FA00,  # .. 0x1FA6F ; Chess Symbols
    0x1FA70,  # .. 0x1FAFF ; Symbols and Pictographs Extended-A
    0x1FB00,  # .. 0x1FBFF ; Symbols for Legacy Computing
    0x1FC00,  # .. 0x1FFFF ; No_Block
    0x20000,  # .. 0x2A6DF ; CJK Unified Ideographs Extension B
    0x2A6E0,  # .. 0x2A6FF ; No_Block
    0x2A700,  # .. 0x2B73F ; CJK Unified Ideographs Extension C
    0x2B740,  # .. 0x2B81F ; CJK Unified Ideographs Extension D
    0x2B820,  # .. 0x2CEAF ; CJK Unified Ideographs Extension E
    0x2CEB0,  # .. 0x2EBEF ; CJK Unified Ideographs Extension F
    0x2EBF0,  # .. 0x2F7FF ; No_Block
    0x2F800,  # .. 0x2FA1F ; CJK Compatibility Ideographs Supplement
    0x2FA20,  # .. 0x2FFFF ; No_Block
    0x30000,  # .. 0x3134F ; CJK Unified Ideographs Extension G
    0x31350,  # .. 0xDFFFF ; No_Block
    0xE0000,  # .. 0xE007F ; Tags
    0xE0080,  # .. 0xE00FF ; No_Block
    0xE0100,  # .. 0xE01EF ; Variation Selectors Supplement
    0xE01F0,  # .. 0xEFFFF ; No_Block
    0xF0000,  # .. 0xFFFFF ; Supplementary Private Use Area-A
    0x100000,  # .. 0x10FFFF ; Supplementary Private Use Area-B
]

VALUES = [
    'Basic Latin',                                     # 0000..007F
    'Latin-1 Supplement',                              # 0080..00FF
    'Latin Extended-A',                                # 0100..017F
    'Latin Extended-B',                                # 0180..024F
    'IPA Extensions',                                  # 0250..02AF
    'Spacing Modifier Letters',                        # 02B0..02FF
    'Combining Diacritical Marks',                     # 0300..036F
    'Greek and Coptic',                                # 0370..03FF
    'Cyrillic',                                        # 0400..04FF
    'Cyrillic Supplement',                             # 0500..052F
    'Armenian',                                        # 0530..058F
    'Hebrew',                                          # 0590..05FF
    'Arabic',                                          # 0600..06FF
    'Syriac',                                          # 0700..074F
    'Arabic Supplement',                               # 0750..077F
    'Thaana',                                          # 0780..07BF
    'NKo',                                             # 07C0..07FF
    'Samaritan',                                       # 0800..083F
    'Mandaic',                                         # 0840..085F
    'Syriac Supplement',                               # 0860..086F
    'No_Block',                                        # 0870..089F
    'Arabic Extended-A',                               # 08A0..08FF
    'Devanagari',                                      # 0900..097F
    'Bengali',                                         # 0980..09FF
    'Gurmukhi',                                        # 0A00..0A7F
    'Gujarati',                                        # 0A80..0AFF
    'Oriya',                                           # 0B00..0B7F
    'Tamil',                                           # 0B80..0BFF
    'Telugu',                                          # 0C00..0C7F
    'Kannada',                                         # 0C80..0CFF
    'Malayalam',                                       # 0D00..0D7F
    'Sinhala',                                         # 0D80..0DFF
    'Thai',                                            # 0E00..0E7F
    'Lao',                                             # 0E80..0EFF
    'Tibetan',                                         # 0F00..0FFF
    'Myanmar',                                         # 1000..109F
    'Georgian',                                        # 10A0..10FF
    'Hangul Jamo',                                     # 1100..11FF
    'Ethiopic',                                        # 1200..137F
    'Ethiopic Supplement',                             # 1380..139F
    'Cherokee',                                        # 13A0..13FF
    'Unified Canadian Aboriginal Syllabics',           # 1400..167F
    'Ogham',                                           # 1680..169F
    'Runic',                                           # 16A0..16FF
    'Tagalog',                                         # 1700..171F
    'Hanunoo',                                         # 1720..173F
    'Buhid',                                           # 1740..175F
    'Tagbanwa',                                        # 1760..177F
    'Khmer',                                           # 1780..17FF
    'Mongolian',                                       # 1800..18AF
    'Unified Canadian Aboriginal Syllabics Extended',  # 18B0..18FF
    'Limbu',                                           # 1900..194F
    'Tai Le',                                          # 1950..197F
    'New Tai Lue',                                     # 1980..19DF
    'Khmer Symbols',                                   # 19E0..19FF
    'Buginese',                                        # 1A00..1A1F
    'Tai Tham',                                        # 1A20..1AAF
    'Combining Diacritical Marks Extended',            # 1AB0..1AFF
    'Balinese',                                        # 1B00..1B7F
    'Sundanese',                                       # 1B80..1BBF
    'Batak',                                           # 1BC0..1BFF
    'Lepcha',                                          # 1C00..1C4F
    'Ol Chiki',                                        # 1C50..1C7F
    'Cyrillic Extended-C',                             # 1C80..1C8F
    'Georgian Extended',                               # 1C90..1CBF
    'Sundanese Supplement',                            # 1CC0..1CCF
    'Vedic Extensions',                                # 1CD0..1CFF
    'Phonetic Extensions',                             # 1D00..1D7F
    'Phonetic Extensions Supplement',                  # 1D80..1DBF
    'Combining Diacritical Marks Supplement',          # 1DC0..1DFF
    'Latin Extended Additional',                       # 1E00..1EFF
    'Greek Extended',                                  # 1F00..1FFF
    'General Punctuation',                             # 2000..206F
    'Superscripts and Subscripts',                     # 2070..209F
    'Currency Symbols',                                # 20A0..20CF
    'Combining Diacritical Marks for Symbols',         # 20D0..20FF
    'Letterlike Symbols',                              # 2100..214F
    'Number Forms',                                    # 2150..218F
    'Arrows',                                          # 2190..21FF
    'Mathematical Operators',                          # 2200..22FF
    'Miscellaneous Technical',                         # 2300..23FF
    'Control Pictures',                                # 2400..243F
    'Optical Character Recognition',                   # 2440..245F
    'Enclosed Alphanumerics',                          # 2460..24FF
    'Box Drawing',                                     # 2500..257F
    'Block Elements',                                  # 2580..259F
    'Geometric Shapes',                                # 25A0..25FF
    'Miscellaneous Symbols',                           # 2600..26FF
    'Dingbats',                                        # 2700..27BF
    'Miscellaneous Mathematical Symbols-A',            # 27C0..27EF
    'Supplemental Arrows-A',                           # 27F0..27FF
    'Braille Patterns',                                # 2800..28FF
    'Supplemental Arrows-B',                           # 2900..297F
    'Miscellaneous Mathematical Symbols-B',            # 2980..29FF
    'Supplemental Mathematical Operators',             # 2A00..2AFF
    'Miscellaneous Symbols and Arrows',                # 2B00..2BFF
    'Glagolitic',                                      # 2C00..2C5F
    'Latin Extended-C',                                # 2C60..2C7F
    'Coptic',                                          # 2C80..2CFF
    'Georgian Supplement',                             # 2D00..2D2F
    'Tifinagh',                                        # 2D30..2D7F
    'Ethiopic Extended',                               # 2D80..2DDF
    'Cyrillic Extended-A',                             # 2DE0..2DFF
    'Supplemental Punctuation',                        # 2E00..2E7F
    'CJK Radicals Supplement',                         # 2E80..2EFF
    'Kangxi Radicals',                                 # 2F00..2FDF
    'No_Block',                                        # 2FE0..2FEF
    'Ideographic Description Characters',              # 2FF0..2FFF
    'CJK Symbols and Punctuation',                     # 3000..303F
    'Hiragana',                                        # 3040..309F
    'Katakana',                                        # 30A0..30FF
    'Bopomofo',                                        # 3100..312F
    'Hangul Compatibility Jamo',                       # 3130..318F
    'Kanbun',                                          # 3190..319F
    'Bopomofo Extended',                               # 31A0..31BF
    'CJK Strokes',                                     # 31C0..31EF
    'Katakana Phonetic Extensions',                    # 31F0..31FF
    'Enclosed CJK Letters and Months',                 # 3200..32FF
    'CJK Compatibility',                               # 3300..33FF
    'CJK Unified Ideographs Extension A',              # 3400..4DBF
    'Yijing Hexagram Symbols',                         # 4DC0..4DFF
    'CJK Unified Ideographs',                          # 4E00..9FFF
    'Yi Syllables',                                    # A000..A48F
    'Yi Radicals',                                     # A490..A4CF
    'Lisu',                                            # A4D0..A4FF
    'Vai',                                             # A500..A63F
    'Cyrillic Extended-B',                             # A640..A69F
    'Bamum',                                           # A6A0..A6FF
    'Modifier Tone Letters',                           # A700..A71F
    'Latin Extended-D',                                # A720..A7FF
    'Syloti Nagri',                                    # A800..A82F
    'Common Indic Number Forms',                       # A830..A83F
    'Phags-pa',                                        # A840..A87F
    'Saurashtra',                                      # A880..A8DF
    'Devanagari Extended',                             # A8E0..A8FF
    'Kayah Li',                                        # A900..A92F
    'Rejang',                                          # A930..A95F
    'Hangul Jamo Extended-A',                          # A960..A97F
    'Javanese',                                        # A980..A9DF
    'Myanmar Extended-B',                              # A9E0..A9FF
    'Cham',                                            # AA00..AA5F
    'Myanmar Extended-A',                              # AA60..AA7F
    'Tai Viet',                                        # AA80..AADF
    'Meetei Mayek Extensions',                         # AAE0..AAFF
    'Ethiopic Extended-A',                             # AB00..AB2F
    'Latin Extended-E',                                # AB30..AB6F
    'Cherokee Supplement',                             # AB70..ABBF
    'Meetei Mayek',                                    # ABC0..ABFF
    'Hangul Syllables',                                # AC00..D7AF
    'Hangul Jamo Extended-B',                          # D7B0..D7FF
    'High Surrogates',                                 # D800..DB7F
    'High Private Use Surrogates',                     # DB80..DBFF
    'Low Surrogates',                                  # DC00..DFFF
    'Private Use Area',                                # E000..F8FF
    'CJK Compatibility Ideographs',                    # F900..FAFF
    'Alphabetic Presentation Forms',                   # FB00..FB4F
    'Arabic Presentation Forms-A',                     # FB50..FDFF
    'Variation Selectors',                             # FE00..FE0F
    'Vertical Forms',                                  # FE10..FE1F
    'Combining Half Marks',                            # FE20..FE2F
    'CJK Compatibility Forms',                         # FE30..FE4F
    'Small Form Variants',                             # FE50..FE6F
    'Arabic Presentation Forms-B',                     # FE70..FEFF
    'Halfwidth and Fullwidth Forms',                   # FF00..FFEF
    'Specials',                                        # FFF0..FFFF
    'Linear B Syllabary',                              # 10000..1007F
    'Linear B Ideograms',                              # 10080..100FF
    'Aegean Numbers',                                  # 10100..1013F
    'Ancient Greek Numbers',                           # 10140..1018F
    'Ancient Symbols',                                 # 10190..101CF
    'Phaistos Disc',                                   # 101D0..101FF
    'No_Block',                                        # 10200..1027F
    'Lycian',                                          # 10280..1029F
    'Carian',                                          # 102A0..102DF
    'Coptic Epact Numbers',                            # 102E0..102FF
    'Old Italic',                                      # 10300..1032F
    'Gothic',                                          # 10330..1034F
    'Old Permic',                                      # 10350..1037F
    'Ugaritic',                                        # 10380..1039F
    'Old Persian',                                     # 103A0..103DF
    'No_Block',                                        # 103E0..103FF
    'Deseret',                                         # 10400..1044F
    'Shavian',                                         # 10450..1047F
    'Osmanya',                                         # 10480..104AF
    'Osage',                                           # 104B0..104FF
    'Elbasan',                                         # 10500..1052F
    'Caucasian Albanian',                              # 10530..1056F
    'No_Block',                                        # 10570..105FF
    'Linear A',                                        # 10600..1077F
    'No_Block',                                        # 10780..107FF
    'Cypriot Syllabary',                               # 10800..1083F
    'Imperial Aramaic',                                # 10840..1085F
    'Palmyrene',                                       # 10860..1087F
    'Nabataean',                                       # 10880..108AF
    'No_Block',                                        # 108B0..108DF
    'Hatran',                                          # 108E0..108FF
    'Phoenician',                                      # 10900..1091F
    'Lydian',                                          # 10920..1093F
    'No_Block',                                        # 10940..1097F
    'Meroitic Hieroglyphs',                            # 10980..1099F
    'Meroitic Cursive',                                # 109A0..109FF
    'Kharoshthi',                                      # 10A00..10A5F
    'Old South Arabian',                               # 10A60..10A7F
    'Old North Arabian',                               # 10A80..10A9F
    'No_Block',                                        # 10AA0..10ABF
    'Manichaean',                                      # 10AC0..10AFF
    'Avestan',                                         # 10B00..10B3F
    'Inscriptional Parthian',                          # 10B40..10B5F
    'Inscriptional Pahlavi',                           # 10B60..10B7F
    'Psalter Pahlavi',                                 # 10B80..10BAF
    'No_Block',                                        # 10BB0..10BFF
    'Old Turkic',                                      # 10C00..10C4F
    'No_Block',                                        # 10C50..10C7F
    'Old Hungarian',                                   # 10C80..10CFF
    'Hanifi Rohingya',                                 # 10D00..10D3F
    'No_Block',                                        # 10D40..10E5F
    'Rumi Numeral Symbols',                            # 10E60..10E7F
    'Yezidi',                                          # 10E80..10EBF
    'No_Block',                                        # 10EC0..10EFF
    'Old Sogdian',                                     # 10F00..10F2F
    'Sogdian',                                         # 10F30..10F6F
    'No_Block',                                        # 10F70..10FAF
    'Chorasmian',                                      # 10FB0..10FDF
    'Elymaic',                                         # 10FE0..10FFF
    'Brahmi',                                          # 11000..1107F
    'Kaithi',                                          # 11080..110CF
    'Sora Sompeng',                                    # 110D0..110FF
    'Chakma',                                          # 11100..1114F
    'Mahajani',                                        # 11150..1117F
    'Sharada',                                         # 11180..111DF
    'Sinhala Archaic Numbers',                         # 111E0..111FF
    'Khojki',                                          # 11200..1124F
    'No_Block',                                        # 11250..1127F
    'Multani',                                         # 11280..112AF
    'Khudawadi',                                       # 112B0..112FF
    'Grantha',                                         # 11300..1137F
    'No_Block',                                        # 11380..113FF
    'Newa',                                            # 11400..1147F
    'Tirhuta',                                         # 11480..114DF
    'No_Block',                                        # 114E0..1157F
    'Siddham',                                         # 11580..115FF
    'Modi',                                            # 11600..1165F
    'Mongolian Supplement',                            # 11660..1167F
    'Takri',                                           # 11680..116CF
    'No_Block',                                        # 116D0..116FF
    'Ahom',                                            # 11700..1173F
    'No_Block',                                        # 11740..117FF
    'Dogra',                                           # 11800..1184F
    'No_Block',                                        # 11850..1189F
    'Warang Citi',                                     # 118A0..118FF
    'Dives Akuru',                                     # 11900..1195F
    'No_Block',                                        # 11960..1199F
    'Nandinagari',                                     # 119A0..119FF
    'Zanabazar Square',                                # 11A00..11A4F
    'Soyombo',                                         # 11A50..11AAF
    'No_Block',                                        # 11AB0..11ABF
    'Pau Cin Hau',                                     # 11AC0..11AFF
    'No_Block',                                        # 11B00..11BFF
    'Bhaiksuki',                                       # 11C00..11C6F
    'Marchen',                                         # 11C70..11CBF
    'No_Block',                                        # 11CC0..11CFF
    'Masaram Gondi',                                   # 11D00..11D5F
    'Gunjala Gondi',                                   # 11D60..11DAF
    'No_Block',                                        # 11DB0..11EDF
    'Makasar',                                         # 11EE0..11EFF
    'No_Block',                                        # 11F00..11FAF
    'Lisu Supplement',                                 # 11FB0..11FBF
    'Tamil Supplement',                                # 11FC0..11FFF
    'Cuneiform',                                       # 12000..123FF
    'Cuneiform Numbers and Punctuation',               # 12400..1247F
    'Early Dynastic Cuneiform',                        # 12480..1254F
    'No_Block',                                        # 12550..12FFF
    'Egyptian Hieroglyphs',                            # 13000..1342F
    'Egyptian Hieroglyph Format Controls',             # 13430..1343F
    'No_Block',                                        # 13440..143FF
    'Anatolian Hieroglyphs',                           # 14400..1467F
    'No_Block',                                        # 14680..167FF
    'Bamum Supplement',                                # 16800..16A3F
    'Mro',                                             # 16A40..16A6F
    'No_Block',                                        # 16A70..16ACF
    'Bassa Vah',                                       # 16AD0..16AFF
    'Pahawh Hmong',                                    # 16B00..16B8F
    'No_Block',                                        # 16B90..16E3F
    'Medefaidrin',                                     # 16E40..16E9F
    'No_Block',                                        # 16EA0..16EFF
    'Miao',                                            # 16F00..16F9F
    'No_Block',                                        # 16FA0..16FDF
    'Ideographic Symbols and Punctuation',             # 16FE0..16FFF
    'Tangut',                                          # 17000..187FF
    'Tangut Components',                               # 18800..18AFF
    'Khitan Small Script',                             # 18B00..18CFF
    'Tangut Supplement',                               # 18D00..18D8F
    'No_Block',                                        # 18D90..1AFFF
    'Kana Supplement',                                 # 1B000..1B0FF
    'Kana Extended-A',                                 # 1B100..1B12F
    'Small Kana Extension',                            # 1B130..1B16F
    'Nushu',                                           # 1B170..1B2FF
    'No_Block',                                        # 1B300..1BBFF
    'Duployan',                                        # 1BC00..1BC9F
    'Shorthand Format Controls',                       # 1BCA0..1BCAF
    'No_Block',                                        # 1BCB0..1CFFF
    'Byzantine Musical Symbols',                       # 1D000..1D0FF
    'Musical Symbols',                                 # 1D100..1D1FF
    'Ancient Greek Musical Notation',                  # 1D200..1D24F
    'No_Block',                                        # 1D250..1D2DF
    'Mayan Numerals',                                  # 1D2E0..1D2FF
    'Tai Xuan Jing Symbols',                           # 1D300..1D35F
    'Counting Rod Numerals',                           # 1D360..1D37F
    'No_Block',                                        # 1D380..1D3FF
    'Mathematical Alphanumeric Symbols',               # 1D400..1D7FF
    'Sutton SignWriting',                              # 1D800..1DAAF
    'No_Block',                                        # 1DAB0..1DFFF
    'Glagolitic Supplement',                           # 1E000..1E02F
    'No_Block',                                        # 1E030..1E0FF
    'Nyiakeng Puachue Hmong',                          # 1E100..1E14F
    'No_Block',                                        # 1E150..1E2BF
    'Wancho',                                          # 1E2C0..1E2FF
    'No_Block',                                        # 1E300..1E7FF
    'Mende Kikakui',                                   # 1E800..1E8DF
    'No_Block',                                        # 1E8E0..1E8FF
    'Adlam',                                           # 1E900..1E95F
    'No_Block',                                        # 1E960..1EC6F
    'Indic Siyaq Numbers',                             # 1EC70..1ECBF
    'No_Block',                                        # 1ECC0..1ECFF
    'Ottoman Siyaq Numbers',                           # 1ED00..1ED4F
    'No_Block',                                        # 1ED50..1EDFF
    'Arabic Mathematical Alphabetic Symbols',          # 1EE00..1EEFF
    'No_Block',                                        # 1EF00..1EFFF
    'Mahjong Tiles',                                   # 1F000..1F02F
    'Domino Tiles',                                    # 1F030..1F09F
    'Playing Cards',                                   # 1F0A0..1F0FF
    'Enclosed Alphanumeric Supplement',                # 1F100..1F1FF
    'Enclosed Ideographic Supplement',                 # 1F200..1F2FF
    'Miscellaneous Symbols and Pictographs',           # 1F300..1F5FF
    'Emoticons',                                       # 1F600..1F64F
    'Ornamental Dingbats',                             # 1F650..1F67F
    'Transport and Map Symbols',                       # 1F680..1F6FF
    'Alchemical Symbols',                              # 1F700..1F77F
    'Geometric Shapes Extended',                       # 1F780..1F7FF
    'Supplemental Arrows-C',                           # 1F800..1F8FF
    'Supplemental Symbols and Pictographs',            # 1F900..1F9FF
    'Chess Symbols',                                   # 1FA00..1FA6F
    'Symbols and Pictographs Extended-A',              # 1FA70..1FAFF
    'Symbols for Legacy Computing',                    # 1FB00..1FBFF
    'No_Block',                                        # 1FC00..1FFFF
    'CJK Unified Ideographs Extension B',              # 20000..2A6DF
    'No_Block',                                        # 2A6E0..2A6FF
    'CJK Unified Ideographs Extension C',              # 2A700..2B73F
    'CJK Unified Ideographs Extension D',              # 2B740..2B81F
    'CJK Unified Ideographs Extension E',              # 2B820..2CEAF
    'CJK Unified Ideographs Extension F',              # 2CEB0..2EBEF
    'No_Block',                                        # 2EBF0..2F7FF
    'CJK Compatibility Ideographs Supplement',         # 2F800..2FA1F
    'No_Block',                                        # 2FA20..2FFFF
    'CJK Unified Ideographs Extension G',              # 30000..3134F
    'No_Block',                                        # 31350..DFFFF
    'Tags',                                            # E0000..E007F
    'No_Block',                                        # E0080..E00FF
    'Variation Selectors Supplement',                  # E0100..E01EF
    'No_Block',                                        # E01F0..EFFFF
    'Supplementary Private Use Area-A',                # F0000..FFFFF
    'Supplementary Private Use Area-B',                # 100000..10FFFF
]



from . import DefaultTable
import sys
import array
import logging


log = logging.getLogger(__name__)


class table__l_o_c_a(DefaultTable.DefaultTable):

	dependencies = ['glyf']

	def decompile(self, data, ttFont):
		longFormat = ttFont['head'].indexToLocFormat
		if longFormat:
			format = "I"
		else:
			format = "H"
		locations = array.array(format)
		locations.frombytes(data)
		if sys.byteorder != "big": locations.byteswap()
		if not longFormat:
			l = array.array("I")
			for i in range(len(locations)):
				l.append(locations[i] * 2)
			locations = l
		if len(locations) < (ttFont['maxp'].numGlyphs + 1):
			log.warning("corrupt 'loca' table, or wrong numGlyphs in 'maxp': %d %d",
				len(locations) - 1, ttFont['maxp'].numGlyphs)
		self.locations = locations

	def compile(self, ttFont):
		try:
			max_location = max(self.locations)
		except AttributeError:
			self.set([])
			max_location = 0
		if max_location < 0x20000 and all(l % 2 == 0 for l in self.locations):
			locations = array.array("H")
			for i in range(len(self.locations)):
				locations.append(self.locations[i] // 2)
			ttFont['head'].indexToLocFormat = 0
		else:
			locations = array.array("I", self.locations)
			ttFont['head'].indexToLocFormat = 1
		if sys.byteorder != "big": locations.byteswap()
		return locations.tobytes()

	def set(self, locations):
		self.locations = array.array("I", locations)

	def toXML(self, writer, ttFont):
		writer.comment("The 'loca' table will be calculated by the compiler")
		writer.newline()

	def __getitem__(self, index):
		return self.locations[index]

	def __len__(self):
		return len(self.locations)



from fontTools.misc.py23 import byteord
from fontTools.misc import sstruct
from fontTools.misc.fixedTools import floatToFixedToStr
from fontTools.misc.textTools import safeEval
# from itertools import *
from . import DefaultTable
from . import grUtils
from array import array
from functools import reduce
import struct, re, sys

Silf_hdr_format = '''
    >
    version:            16.16F
'''

Silf_hdr_format_3 = '''
    >
    version:            16.16F
    compilerVersion:    L
    numSilf:            H
                        x
                        x
'''

Silf_part1_format_v3 = '''
    >
    ruleVersion:        16.16F
    passOffset:         H
    pseudosOffset:      H
'''

Silf_part1_format = '''
    >
    maxGlyphID:         H
    extraAscent:        h
    extraDescent:       h
    numPasses:          B
    iSubst:             B
    iPos:               B
    iJust:              B
    iBidi:              B
    flags:              B
    maxPreContext:      B
    maxPostContext:     B
    attrPseudo:         B
    attrBreakWeight:    B
    attrDirectionality: B
    attrMirroring:      B
    attrSkipPasses:     B
    numJLevels:         B
'''

Silf_justify_format = '''
    >
    attrStretch:        B
    attrShrink:         B
    attrStep:           B
    attrWeight:         B
    runto:              B
                        x
                        x
                        x
'''

Silf_part2_format = '''
    >
    numLigComp:         H
    numUserDefn:        B
    maxCompPerLig:      B
    direction:          B
    attCollisions:      B
                        x
                        x
                        x
    numCritFeatures:    B
'''

Silf_pseudomap_format = '''
    >
    unicode:            L
    nPseudo:            H
'''

Silf_pseudomap_format_h = '''
    >
    unicode:            H
    nPseudo:            H
'''

Silf_classmap_format = '''
    >
    numClass:           H
    numLinear:          H
'''

Silf_lookupclass_format = '''
    >
    numIDs:             H
    searchRange:        H
    entrySelector:      H
    rangeShift:         H
'''

Silf_lookuppair_format = '''
    >
    glyphId:            H
    index:              H
'''

Silf_pass_format = '''
    >
    flags:              B
    maxRuleLoop:        B
    maxRuleContext:     B
    maxBackup:          B
    numRules:           H
    fsmOffset:          H
    pcCode:             L
    rcCode:             L
    aCode:              L
    oDebug:             L
    numRows:            H
    numTransitional:    H
    numSuccess:         H
    numColumns:         H
'''

aCode_info = (
    ("NOP", 0),
    ("PUSH_BYTE", "b"),
    ("PUSH_BYTE_U", "B"),
    ("PUSH_SHORT", ">h"),
    ("PUSH_SHORT_U", ">H"),
    ("PUSH_LONG", ">L"),
    ("ADD", 0),
    ("SUB", 0),
    ("MUL", 0),
    ("DIV", 0),
    ("MIN", 0),
    ("MAX", 0),
    ("NEG", 0),
    ("TRUNC8", 0),
    ("TRUNC16", 0),
    ("COND", 0),
    ("AND", 0),         # x10
    ("OR", 0),
    ("NOT", 0),
    ("EQUAL", 0),
    ("NOT_EQ", 0),
    ("LESS", 0),
    ("GTR", 0),
    ("LESS_EQ", 0),
    ("GTR_EQ", 0),
    ("NEXT", 0),
    ("NEXT_N", "b"),
    ("COPY_NEXT", 0),
    ("PUT_GLYPH_8BIT_OBS", "B"),
    ("PUT_SUBS_8BIT_OBS", "bBB"),
    ("PUT_COPY", "b"),
    ("INSERT", 0),
    ("DELETE", 0),      # x20
    ("ASSOC", -1),
    ("CNTXT_ITEM", "bB"),
    ("ATTR_SET", "B"),
    ("ATTR_ADD", "B"),
    ("ATTR_SUB", "B"),
    ("ATTR_SET_SLOT", "B"),
    ("IATTR_SET_SLOT", "BB"),
    ("PUSH_SLOT_ATTR", "Bb"),
    ("PUSH_GLYPH_ATTR_OBS", "Bb"),
    ("PUSH_GLYPH_METRIC", "Bbb"),
    ("PUSH_FEAT", "Bb"),
    ("PUSH_ATT_TO_GATTR_OBS", "Bb"),
    ("PUSH_ATT_TO_GLYPH_METRIC", "Bbb"),
    ("PUSH_ISLOT_ATTR", "Bbb"),
    ("PUSH_IGLYPH_ATTR", "Bbb"),
    ("POP_RET", 0),     # x30
    ("RET_ZERO", 0),
    ("RET_TRUE", 0),
    ("IATTR_SET", "BB"),
    ("IATTR_ADD", "BB"),
    ("IATTR_SUB", "BB"),
    ("PUSH_PROC_STATE", "B"),
    ("PUSH_VERSION", 0),
    ("PUT_SUBS", ">bHH"),
    ("PUT_SUBS2", 0),
    ("PUT_SUBS3", 0),
    ("PUT_GLYPH", ">H"),
    ("PUSH_GLYPH_ATTR", ">Hb"),
    ("PUSH_ATT_TO_GLYPH_ATTR", ">Hb"),
    ("BITOR", 0),
    ("BITAND", 0),
    ("BITNOT", 0),      # x40
    ("BITSET", ">HH"),
    ("SET_FEAT", "Bb")
)
aCode_map = dict([(x[0], (i, x[1])) for i,x in enumerate(aCode_info)])

def disassemble(aCode):
    codelen = len(aCode)
    pc = 0
    res = []
    while pc < codelen:
        opcode = byteord(aCode[pc:pc+1])
        if opcode > len(aCode_info):
            instr = aCode_info[0]
        else:
            instr = aCode_info[opcode]
        pc += 1
        if instr[1] != 0 and pc >= codelen : return res
        if instr[1] == -1:
            count = byteord(aCode[pc])
            fmt = "%dB" % count
            pc += 1
        elif instr[1] == 0:
            fmt = ""
        else :
            fmt = instr[1]
        if fmt == "":
            res.append(instr[0])
            continue
        parms = struct.unpack_from(fmt, aCode[pc:])
        res.append(instr[0] + "(" + ", ".join(map(str, parms)) + ")")
        pc += struct.calcsize(fmt)
    return res

instre = re.compile(r"^\s*([^(]+)\s*(?:\(([^)]+)\))?")
def assemble(instrs):
    res = b""
    for inst in instrs:
        m = instre.match(inst)
        if not m or not m.group(1) in aCode_map:
            continue
        opcode, parmfmt = aCode_map[m.group(1)]
        res += struct.pack("B", opcode)
        if m.group(2):
            if parmfmt == 0:
                continue
            parms = [int(x) for x in re.split(r",\s*", m.group(2))]
            if parmfmt == -1:
                l = len(parms)
                res += struct.pack(("%dB" % (l+1)), l, *parms)
            else:
                res += struct.pack(parmfmt, *parms)
    return res

def writecode(tag, writer, instrs):
    writer.begintag(tag)
    writer.newline()
    for l in disassemble(instrs):
        writer.write(l)
        writer.newline()
    writer.endtag(tag)
    writer.newline()

def readcode(content):
    res = []
    for e in content_string(content).split('\n'):
        e = e.strip()
        if not len(e): continue
        res.append(e)
    return assemble(res)
    
attrs_info=('flags', 'extraAscent', 'extraDescent', 'maxGlyphID',
            'numLigComp', 'numUserDefn', 'maxCompPerLig', 'direction', 'lbGID')
attrs_passindexes = ('iSubst', 'iPos', 'iJust', 'iBidi')
attrs_contexts = ('maxPreContext', 'maxPostContext')
attrs_attributes = ('attrPseudo', 'attrBreakWeight', 'attrDirectionality',
                    'attrMirroring', 'attrSkipPasses', 'attCollisions')
pass_attrs_info = ('flags', 'maxRuleLoop', 'maxRuleContext', 'maxBackup',
            'minRulePreContext', 'maxRulePreContext', 'collisionThreshold')
pass_attrs_fsm = ('numRows', 'numTransitional', 'numSuccess', 'numColumns')

def writesimple(tag, self, writer, *attrkeys):
    attrs = dict([(k, getattr(self, k)) for k in attrkeys])
    writer.simpletag(tag, **attrs)
    writer.newline()

def getSimple(self, attrs, *attr_list):
    for k in attr_list:
        if k in attrs:
            setattr(self, k, int(safeEval(attrs[k])))

def content_string(contents):
    res = ""
    for element in contents:
        if isinstance(element, tuple): continue
        res += element
    return res.strip()

def wrapline(writer, dat, length=80):
    currline = ""
    for d in dat:
        if len(currline) > length:
            writer.write(currline[:-1])
            writer.newline()
            currline = ""
        currline += d + " "
    if len(currline):
        writer.write(currline[:-1])
        writer.newline()

class _Object() :
    pass

class table_S__i_l_f(DefaultTable.DefaultTable):
    '''Silf table support'''

    def __init__(self, tag=None):
        DefaultTable.DefaultTable.__init__(self, tag)
        self.silfs = []

    def decompile(self, data, ttFont):
        sstruct.unpack2(Silf_hdr_format, data, self)
        self.version = float(floatToFixedToStr(self.version, precisionBits=16))
        if self.version >= 5.0:
            (data, self.scheme) = grUtils.decompress(data)
            sstruct.unpack2(Silf_hdr_format_3, data, self)
            base = sstruct.calcsize(Silf_hdr_format_3)
        elif self.version < 3.0:
            self.numSilf = struct.unpack('>H', data[4:6])
            self.scheme = 0
            self.compilerVersion = 0
            base = 8
        else:
            self.scheme = 0
            sstruct.unpack2(Silf_hdr_format_3, data, self)
            base = sstruct.calcsize(Silf_hdr_format_3)

        silfoffsets = struct.unpack_from(('>%dL' % self.numSilf), data[base:])
        for offset in silfoffsets:
            s = Silf()
            self.silfs.append(s)
            s.decompile(data[offset:], ttFont, self.version)

    def compile(self, ttFont):
        self.numSilf = len(self.silfs)
        if self.version < 3.0:
            hdr = sstruct.pack(Silf_hdr_format, self)
            hdr += struct.pack(">HH", self.numSilf, 0)
        else:
            hdr = sstruct.pack(Silf_hdr_format_3, self)
        offset = len(hdr) + 4 * self.numSilf
        data = b""
        for s in self.silfs:
            hdr += struct.pack(">L", offset)
            subdata = s.compile(ttFont, self.version)
            offset += len(subdata)
            data += subdata
        if self.version >= 5.0:
            return grUtils.compress(self.scheme, hdr+data)
        return hdr+data

    def toXML(self, writer, ttFont):
        writer.comment('Attributes starting with _ are informative only')
        writer.newline()
        writer.simpletag('version', version=self.version,
            compilerVersion=self.compilerVersion, compressionScheme=self.scheme)
        writer.newline()
        for s in self.silfs:
            writer.begintag('silf')
            writer.newline()
            s.toXML(writer, ttFont, self.version)
            writer.endtag('silf')
            writer.newline()

    def fromXML(self, name, attrs, content, ttFont):
        if name == 'version':
            self.scheme=int(safeEval(attrs['compressionScheme']))
            self.version = float(safeEval(attrs['version']))
            self.compilerVersion = int(safeEval(attrs['compilerVersion']))
            return
        if name == 'silf':
            s = Silf()
            self.silfs.append(s)
            for element in content:
                if not isinstance(element, tuple): continue
                tag, attrs, subcontent = element
                s.fromXML(tag, attrs, subcontent, ttFont, self.version)

class Silf(object):
    '''A particular Silf subtable'''

    def __init__(self):
        self.passes = []
        self.scriptTags = []
        self.critFeatures = []
        self.jLevels = []
        self.pMap = {}

    def decompile(self, data, ttFont, version=2.0):
        if version >= 3.0 :
            _, data = sstruct.unpack2(Silf_part1_format_v3, data, self)
            self.ruleVersion = float(floatToFixedToStr(self.ruleVersion, precisionBits=16))
        _, data = sstruct.unpack2(Silf_part1_format, data, self)
        for jlevel in range(self.numJLevels):
            j, data = sstruct.unpack2(Silf_justify_format, data, _Object())
            self.jLevels.append(j)
        _, data = sstruct.unpack2(Silf_part2_format, data, self)
        if self.numCritFeatures:
            self.critFeatures = struct.unpack_from(('>%dH' % self.numCritFeatures), data)
        data = data[self.numCritFeatures * 2 + 1:]
        (numScriptTag,) = struct.unpack_from('B', data)
        if numScriptTag:
            self.scriptTags = [struct.unpack("4s", data[x:x+4])[0].decode("ascii") for x in range(1, 1 + 4 * numScriptTag, 4)]
        data = data[1 + 4 * numScriptTag:]
        (self.lbGID,) = struct.unpack('>H', data[:2])
        if self.numPasses:
            self.oPasses = struct.unpack(('>%dL' % (self.numPasses+1)), data[2:6+4*self.numPasses])
        data = data[6 + 4 * self.numPasses:]
        (numPseudo,) = struct.unpack(">H", data[:2])
        for i in range(numPseudo):
            if version >= 3.0:
                pseudo = sstruct.unpack(Silf_pseudomap_format, data[8+6*i:14+6*i], _Object())
            else:
                pseudo = sstruct.unpack(Silf_pseudomap_format_h, data[8+4*i:12+4*i], _Object())
            self.pMap[pseudo.unicode] = ttFont.getGlyphName(pseudo.nPseudo)
        data = data[8 + 6 * numPseudo:]
        currpos = (sstruct.calcsize(Silf_part1_format)
                    + sstruct.calcsize(Silf_justify_format) * self.numJLevels
                    + sstruct.calcsize(Silf_part2_format) + 2 * self.numCritFeatures
                    + 1 + 1 + 4 * numScriptTag + 6 + 4 * self.numPasses + 8 + 6 * numPseudo)
        if version >= 3.0:
            currpos += sstruct.calcsize(Silf_part1_format_v3)
        self.classes = Classes()
        self.classes.decompile(data, ttFont, version)
        for i in range(self.numPasses):
            p = Pass()
            self.passes.append(p)
            p.decompile(data[self.oPasses[i]-currpos:self.oPasses[i+1]-currpos],
                        ttFont, version)

    def compile(self, ttFont, version=2.0):
        self.numPasses = len(self.passes)
        self.numJLevels = len(self.jLevels)
        self.numCritFeatures = len(self.critFeatures)
        numPseudo = len(self.pMap)
        data = b""
        if version >= 3.0:
            hdroffset = sstruct.calcsize(Silf_part1_format_v3)
        else:
            hdroffset = 0
        data += sstruct.pack(Silf_part1_format, self)
        for j in self.jLevels:
            data += sstruct.pack(Silf_justify_format, j)
        data += sstruct.pack(Silf_part2_format, self)
        if self.numCritFeatures:
            data += struct.pack((">%dH" % self.numCritFeaturs), *self.critFeatures)
        data += struct.pack("BB", 0, len(self.scriptTags))
        if len(self.scriptTags):
            tdata = [struct.pack("4s", x.encode("ascii")) for x in self.scriptTags]
            data += b"".join(tdata)
        data += struct.pack(">H", self.lbGID)
        self.passOffset = len(data)

        data1 = grUtils.bininfo(numPseudo, 6)
        currpos = hdroffset + len(data) + 4 * (self.numPasses + 1)
        self.pseudosOffset = currpos + len(data1)
        for u, p in sorted(self.pMap.items()):
            data1 += struct.pack((">LH" if version >= 3.0 else ">HH"),
                                u, ttFont.getGlyphID(p))
        data1 += self.classes.compile(ttFont, version)
        currpos += len(data1)
        data2 = b""
        datao = b""
        for i, p in enumerate(self.passes):
            base = currpos + len(data2)
            datao += struct.pack(">L", base)
            data2 += p.compile(ttFont, base, version)
        datao += struct.pack(">L", currpos + len(data2))

        if version >= 3.0:
            data3 = sstruct.pack(Silf_part1_format_v3, self)
        else:
            data3 = b""
        return data3 + data + datao + data1 + data2


    def toXML(self, writer, ttFont, version=2.0):
        if version >= 3.0:
            writer.simpletag('version', ruleVersion=self.ruleVersion)
            writer.newline()
        writesimple('info', self, writer, *attrs_info)
        writesimple('passindexes', self, writer, *attrs_passindexes)
        writesimple('contexts', self, writer, *attrs_contexts)
        writesimple('attributes', self, writer, *attrs_attributes)
        if len(self.jLevels):
            writer.begintag('justifications')
            writer.newline()
            jformat, jnames, jfixes = sstruct.getformat(Silf_justify_format)
            for i, j in enumerate(self.jLevels):
                attrs = dict([(k, getattr(j, k)) for k in jnames])
                writer.simpletag('justify', **attrs)
                writer.newline()
            writer.endtag('justifications')
            writer.newline()
        if len(self.critFeatures):
            writer.begintag('critFeatures')
            writer.newline()
            writer.write(" ".join(map(str, self.critFeatures)))
            writer.newline()
            writer.endtag('critFeatures')
            writer.newline()
        if len(self.scriptTags):
            writer.begintag('scriptTags')
            writer.newline()
            writer.write(" ".join(self.scriptTags))
            writer.newline()
            writer.endtag('scriptTags')
            writer.newline()
        if self.pMap:
            writer.begintag('pseudoMap')
            writer.newline()
            for k, v in sorted(self.pMap.items()):
                writer.simpletag('pseudo', unicode=hex(k), pseudo=v)
                writer.newline()
            writer.endtag('pseudoMap')
            writer.newline()
        self.classes.toXML(writer, ttFont, version)
        if len(self.passes):
            writer.begintag('passes')
            writer.newline()
            for i, p in enumerate(self.passes):
                writer.begintag('pass', _index=i)
                writer.newline()
                p.toXML(writer, ttFont, version)
                writer.endtag('pass')
                writer.newline()
            writer.endtag('passes')
            writer.newline()

    def fromXML(self, name, attrs, content, ttFont, version=2.0):
        if name == 'version':
            self.ruleVersion = float(safeEval(attrs.get('ruleVersion', "0")))
        if name == 'info':
            getSimple(self, attrs, *attrs_info)
        elif name == 'passindexes':
            getSimple(self, attrs, *attrs_passindexes)
        elif name == 'contexts':
            getSimple(self, attrs, *attrs_contexts)
        elif name == 'attributes':
            getSimple(self, attrs, *attrs_attributes)
        elif name == 'justifications':
            for element in content:
                if not isinstance(element, tuple): continue
                (tag, attrs, subcontent) = element
                if tag == 'justify':
                    j = _Object()
                    for k, v in attrs.items():
                        setattr(j, k, int(v))
                    self.jLevels.append(j)
        elif name == 'critFeatures':
            self.critFeatures = []
            element = content_string(content)
            self.critFeatures.extend(map(int, element.split()))
        elif name == 'scriptTags':
            self.scriptTags = []
            element = content_string(content)
            for n in element.split():
                self.scriptTags.append(n)
        elif name == 'pseudoMap':
            self.pMap = {}
            for element in content:
                if not isinstance(element, tuple): continue
                (tag, attrs, subcontent) = element
                if tag == 'pseudo':
                    k = int(attrs['unicode'], 16)
                    v = attrs['pseudo']
                self.pMap[k] = v
        elif name == 'classes':
            self.classes = Classes()
            for element in content:
                if not isinstance(element, tuple): continue
                tag, attrs, subcontent = element
                self.classes.fromXML(tag, attrs, subcontent, ttFont, version)
        elif name == 'passes':
            for element in content:
                if not isinstance(element, tuple): continue
                tag, attrs, subcontent = element
                if tag == 'pass':
                    p = Pass()
                    for e in subcontent:
                        if not isinstance(e, tuple): continue
                        p.fromXML(e[0], e[1], e[2], ttFont, version)
                    self.passes.append(p)


class Classes(object):

    def __init__(self):
        self.linear = []
        self.nonLinear = []

    def decompile(self, data, ttFont, version=2.0):
        sstruct.unpack2(Silf_classmap_format, data, self)
        if version >= 4.0 :
            oClasses = struct.unpack((">%dL" % (self.numClass+1)),
                                        data[4:8+4*self.numClass])
        else:
            oClasses = struct.unpack((">%dH" % (self.numClass+1)),
                                        data[4:6+2*self.numClass])
        for s,e in zip(oClasses[:self.numLinear], oClasses[1:self.numLinear+1]):
            self.linear.append(ttFont.getGlyphName(x) for x in
                                   struct.unpack((">%dH" % ((e-s)/2)), data[s:e]))
        for s,e in zip(oClasses[self.numLinear:self.numClass],
                        oClasses[self.numLinear+1:self.numClass+1]):
            nonLinids = [struct.unpack(">HH", data[x:x+4]) for x in range(s+8, e, 4)]
            nonLin = dict([(ttFont.getGlyphName(x[0]), x[1]) for x in nonLinids])
            self.nonLinear.append(nonLin)

    def compile(self, ttFont, version=2.0):
        data = b""
        oClasses = []
        if version >= 4.0:
            offset = 8 + 4 * (len(self.linear) + len(self.nonLinear))
        else:
            offset = 6 + 2 * (len(self.linear) + len(self.nonLinear))
        for l in self.linear:
            oClasses.append(len(data) + offset)
            gs = [ttFont.getGlyphID(x) for x in l]
            data += struct.pack((">%dH" % len(l)), *gs)
        for l in self.nonLinear:
            oClasses.append(len(data) + offset)
            gs = [(ttFont.getGlyphID(x[0]), x[1]) for x in l.items()]
            data += grUtils.bininfo(len(gs))
            data += b"".join([struct.pack(">HH", *x) for x in sorted(gs)])
        oClasses.append(len(data) + offset)
        self.numClass = len(oClasses) - 1
        self.numLinear = len(self.linear)
        return sstruct.pack(Silf_classmap_format, self) + \
               struct.pack(((">%dL" if version >= 4.0 else ">%dH") % len(oClasses)),
                            *oClasses) + data

    def toXML(self, writer, ttFont, version=2.0):
        writer.begintag('classes')
        writer.newline()
        writer.begintag('linearClasses')
        writer.newline()
        for i,l in enumerate(self.linear):
            writer.begintag('linear', _index=i)
            writer.newline()
            wrapline(writer, l)
            writer.endtag('linear')
            writer.newline()
        writer.endtag('linearClasses')
        writer.newline()
        writer.begintag('nonLinearClasses')
        writer.newline()
        for i, l in enumerate(self.nonLinear):
            writer.begintag('nonLinear', _index=i + self.numLinear)
            writer.newline()
            for inp, ind in l.items():
                writer.simpletag('map', glyph=inp, index=ind)
                writer.newline()
            writer.endtag('nonLinear')
            writer.newline()
        writer.endtag('nonLinearClasses')
        writer.newline()
        writer.endtag('classes')
        writer.newline()

    def fromXML(self, name, attrs, content, ttFont, version=2.0):
        if name == 'linearClasses':
            for element in content:
                if not isinstance(element, tuple): continue
                tag, attrs, subcontent = element
                if tag == 'linear':
                    l = content_string(subcontent).split()
                    self.linear.append(l)
        elif name == 'nonLinearClasses':
            for element in content:
                if not isinstance(element, tuple): continue
                tag, attrs, subcontent = element
                if tag =='nonLinear':
                    l = {}
                    for e in subcontent:
                        if not isinstance(e, tuple): continue
                        tag, attrs, subsubcontent = e
                        if tag == 'map':
                            l[attrs['glyph']] = int(safeEval(attrs['index']))
                    self.nonLinear.append(l)

class Pass(object):

    def __init__(self):
        self.colMap = {}
        self.rules = []
        self.rulePreContexts = []
        self.ruleSortKeys = []
        self.ruleConstraints = []
        self.passConstraints = b""
        self.actions = []
        self.stateTrans = []
        self.startStates = []

    def decompile(self, data, ttFont, version=2.0):
        _, data = sstruct.unpack2(Silf_pass_format, data, self)
        (numRange, _, _, _) = struct.unpack(">4H", data[:8])
        data = data[8:]
        for i in range(numRange):
            (first, last, col) = struct.unpack(">3H", data[6*i:6*i+6])
            for g in range(first, last+1):
                self.colMap[ttFont.getGlyphName(g)] = col
        data = data[6*numRange:]
        oRuleMap = struct.unpack_from((">%dH" % (self.numSuccess + 1)), data)
        data = data[2+2*self.numSuccess:]
        rules = struct.unpack_from((">%dH" % oRuleMap[-1]), data)
        self.rules = [rules[s:e] for (s,e) in zip(oRuleMap, oRuleMap[1:])]
        data = data[2*oRuleMap[-1]:]
        (self.minRulePreContext, self.maxRulePreContext) = struct.unpack('BB', data[:2])
        numStartStates = self.maxRulePreContext - self.minRulePreContext + 1
        self.startStates = struct.unpack((">%dH" % numStartStates),
                                        data[2:2 + numStartStates * 2])
        data = data[2+numStartStates*2:]
        self.ruleSortKeys = struct.unpack((">%dH" % self.numRules), data[:2 * self.numRules])
        data = data[2*self.numRules:]
        self.rulePreContexts = struct.unpack(("%dB" % self.numRules), data[:self.numRules])
        data = data[self.numRules:]
        (self.collisionThreshold, pConstraint) = struct.unpack(">BH", data[:3])
        oConstraints = list(struct.unpack((">%dH" % (self.numRules + 1)),
                                        data[3:5 + self.numRules * 2]))
        data = data[5 + self.numRules * 2:]
        oActions = list(struct.unpack((">%dH" % (self.numRules + 1)),
                                        data[:2 + self.numRules * 2]))
        data = data[2 * self.numRules + 2:]
        for i in range(self.numTransitional):
            a = array("H", data[i*self.numColumns*2:(i+1)*self.numColumns*2])
            if sys.byteorder != "big": a.byteswap()
            self.stateTrans.append(a)
        data = data[self.numTransitional * self.numColumns * 2 + 1:]
        self.passConstraints = data[:pConstraint]
        data = data[pConstraint:]
        for i in range(len(oConstraints)-2,-1,-1):
            if oConstraints[i] == 0 :
                oConstraints[i] = oConstraints[i+1]
        self.ruleConstraints = [(data[s:e] if (e-s > 1) else b"") for (s,e) in zip(oConstraints, oConstraints[1:])]
        data = data[oConstraints[-1]:]
        self.actions = [(data[s:e] if (e-s > 1) else "") for (s,e) in zip(oActions, oActions[1:])]
        data = data[oActions[-1]:]
        # not using debug

    def compile(self, ttFont, base, version=2.0):
        # build it all up backwards
        oActions = reduce(lambda a, x: (a[0]+len(x), a[1]+[a[0]]), self.actions + [b""], (0, []))[1]
        oConstraints = reduce(lambda a, x: (a[0]+len(x), a[1]+[a[0]]), self.ruleConstraints + [b""], (1, []))[1]
        constraintCode = b"\000" + b"".join(self.ruleConstraints)
        transes = []
        for t in self.stateTrans:
            if sys.byteorder != "big": t.byteswap()
            transes.append(t.tobytes())
            if sys.byteorder != "big": t.byteswap()
        if not len(transes):
            self.startStates = [0]
        oRuleMap = reduce(lambda a, x: (a[0]+len(x), a[1]+[a[0]]), self.rules+[[]], (0, []))[1]
        passRanges = []
        gidcolmap = dict([(ttFont.getGlyphID(x[0]), x[1]) for x in self.colMap.items()])
        for e in grUtils.entries(gidcolmap, sameval = True):
            if e[1]:
                passRanges.append((e[0], e[0]+e[1]-1, e[2][0]))
        self.numRules = len(self.actions)
        self.fsmOffset = (sstruct.calcsize(Silf_pass_format) + 8 + len(passRanges) * 6
                    + len(oRuleMap) * 2 + 2 * oRuleMap[-1] + 2
                    + 2 * len(self.startStates) + 3 * self.numRules + 3
                    + 4 * self.numRules + 4)
        self.pcCode = self.fsmOffset + 2*self.numTransitional*self.numColumns + 1 + base
        self.rcCode = self.pcCode + len(self.passConstraints)
        self.aCode = self.rcCode + len(constraintCode)
        self.oDebug = 0
        # now generate output
        data = sstruct.pack(Silf_pass_format, self)
        data += grUtils.bininfo(len(passRanges), 6)
        data += b"".join(struct.pack(">3H", *p) for p in passRanges)
        data += struct.pack((">%dH" % len(oRuleMap)), *oRuleMap)
        flatrules = reduce(lambda a,x: a+x, self.rules, [])
        data += struct.pack((">%dH" % oRuleMap[-1]), *flatrules)
        data += struct.pack("BB", self.minRulePreContext, self.maxRulePreContext)
        data += struct.pack((">%dH" % len(self.startStates)), *self.startStates)
        data += struct.pack((">%dH" % self.numRules), *self.ruleSortKeys)
        data += struct.pack(("%dB" % self.numRules), *self.rulePreContexts)
        data += struct.pack(">BH", self.collisionThreshold, len(self.passConstraints))
        data += struct.pack((">%dH" % (self.numRules+1)), *oConstraints)
        data += struct.pack((">%dH" % (self.numRules+1)), *oActions)
        return data + b"".join(transes) + struct.pack("B", 0) + \
                self.passConstraints + constraintCode + b"".join(self.actions)

    def toXML(self, writer, ttFont, version=2.0):
        writesimple('info', self, writer, *pass_attrs_info)
        writesimple('fsminfo', self, writer, *pass_attrs_fsm)
        writer.begintag('colmap')
        writer.newline()
        wrapline(writer, ["{}={}".format(*x) for x in sorted(self.colMap.items(),
                                        key=lambda x:ttFont.getGlyphID(x[0]))])
        writer.endtag('colmap')
        writer.newline()
        writer.begintag('staterulemap')
        writer.newline()
        for i, r in enumerate(self.rules):
            writer.simpletag('state', number = self.numRows - self.numSuccess + i,
                                rules = " ".join(map(str, r)))
            writer.newline()
        writer.endtag('staterulemap')
        writer.newline()
        writer.begintag('rules')
        writer.newline()
        for i in range(len(self.actions)):
            writer.begintag('rule', index=i, precontext=self.rulePreContexts[i],
                            sortkey=self.ruleSortKeys[i])
            writer.newline()
            if len(self.ruleConstraints[i]):
                writecode('constraint', writer, self.ruleConstraints[i])
            writecode('action', writer, self.actions[i])
            writer.endtag('rule')
            writer.newline()
        writer.endtag('rules')
        writer.newline()
        if len(self.passConstraints):
            writecode('passConstraint', writer, self.passConstraints)
        if len(self.stateTrans):
            writer.begintag('fsm')
            writer.newline()
            writer.begintag('starts')
            writer.write(" ".join(map(str, self.startStates)))
            writer.endtag('starts')
            writer.newline()
            for i, s in enumerate(self.stateTrans):
                writer.begintag('row', _i=i)
                # no newlines here
                writer.write(" ".join(map(str, s)))
                writer.endtag('row')
                writer.newline()
            writer.endtag('fsm')
            writer.newline()

    def fromXML(self, name, attrs, content, ttFont, version=2.0):
        if name == 'info':
            getSimple(self, attrs, *pass_attrs_info)
        elif name == 'fsminfo':
            getSimple(self, attrs, *pass_attrs_fsm)
        elif name == 'colmap':
            e = content_string(content)
            for w in e.split():
                x = w.split('=')
                if len(x) != 2 or x[0] == '' or x[1] == '': continue
                self.colMap[x[0]] = int(x[1])
        elif name == 'staterulemap':
            for e in content:
                if not isinstance(e, tuple): continue
                tag, a, c = e
                if tag == 'state':
                    self.rules.append([int(x) for x in a['rules'].split(" ")])
        elif name == 'rules':
            for element in content:
                if not isinstance(element, tuple): continue
                tag, a, c = element
                if tag != 'rule': continue
                self.rulePreContexts.append(int(a['precontext']))
                self.ruleSortKeys.append(int(a['sortkey']))
                con = b""
                act = b""
                for e in c:
                    if not isinstance(e, tuple): continue
                    tag, a, subc = e
                    if tag == 'constraint':
                        con = readcode(subc)
                    elif tag == 'action':
                        act = readcode(subc)
                self.actions.append(act)
                self.ruleConstraints.append(con)
        elif name == 'passConstraint':
            self.passConstraints = readcode(content)
        elif name == 'fsm':
            for element in content:
                if not isinstance(element, tuple): continue
                tag, a, c = element
                if tag == 'row':
                    s = array('H')
                    e = content_string(c)
                    s.extend(map(int, e.split()))
                    self.stateTrans.append(s)
                elif tag == 'starts':
                    s = []
                    e = content_string(c)
                    s.extend(map(int, e.split()))
                    self.startStates = s




"""
Merge OpenType Layout tables (GDEF / GPOS / GSUB).
"""
import copy
from operator import ior
from fontTools.misc import classifyTools
from fontTools.misc.roundTools import otRound
from fontTools.ttLib.tables import otTables as ot
from fontTools.ttLib.tables import otBase as otBase
from fontTools.ttLib.tables.DefaultTable import DefaultTable
from fontTools.varLib import builder, models, varStore
from fontTools.varLib.models import nonNone, allNone, allEqual, allEqualTo
from fontTools.varLib.varStore import VarStoreInstancer
from functools import reduce
from fontTools.otlLib.builder import buildSinglePos

from .errors import (
    ShouldBeConstant,
    FoundANone,
    MismatchedTypes,
    LengthsDiffer,
    KeysDiffer,
    InconsistentGlyphOrder,
    InconsistentExtensions,
    UnsupportedFormat,
    UnsupportedFormat,
    VarLibMergeError,
)

class Merger(object):

	def __init__(self, font=None):
		self.font = font

	@classmethod
	def merger(celf, clazzes, attrs=(None,)):
		assert celf != Merger, 'Subclass Merger instead.'
		if 'mergers' not in celf.__dict__:
			celf.mergers = {}
		if type(clazzes) == type:
			clazzes = (clazzes,)
		if type(attrs) == str:
			attrs = (attrs,)
		def wrapper(method):
			assert method.__name__ == 'merge'
			done = []
			for clazz in clazzes:
				if clazz in done: continue # Support multiple names of a clazz
				done.append(clazz)
				mergers = celf.mergers.setdefault(clazz, {})
				for attr in attrs:
					assert attr not in mergers, \
						"Oops, class '%s' has merge function for '%s' defined already." % (clazz.__name__, attr)
					mergers[attr] = method
			return None
		return wrapper

	@classmethod
	def mergersFor(celf, thing, _default={}):
		typ = type(thing)

		for celf in celf.mro():

			mergers = getattr(celf, 'mergers', None)
			if mergers is None:
				break;

			m = celf.mergers.get(typ, None)
			if m is not None:
				return m

		return _default

	def mergeObjects(self, out, lst, exclude=()):
		if hasattr(out, "ensureDecompiled"):
			out.ensureDecompiled()
		for item in lst:
			if hasattr(item, "ensureDecompiled"):
				item.ensureDecompiled()
		keys = sorted(vars(out).keys())
		if not all(keys == sorted(vars(v).keys()) for v in lst):
			raise KeysDiffer(self, expected=keys,
				got=[sorted(vars(v).keys()) for v in lst]
			)
		mergers = self.mergersFor(out)
		defaultMerger = mergers.get('*', self.__class__.mergeThings)
		try:
			for key in keys:
				if key in exclude: continue
				value = getattr(out, key)
				values = [getattr(table, key) for table in lst]
				mergerFunc = mergers.get(key, defaultMerger)
				mergerFunc(self, value, values)
		except VarLibMergeError as e:
			e.stack.append('.'+key)
			raise

	def mergeLists(self, out, lst):
		if not allEqualTo(out, lst, len):
			raise LengthsDiffer(self, expected=len(out), got=[len(x) for x in lst])
		for i,(value,values) in enumerate(zip(out, zip(*lst))):
			try:
				self.mergeThings(value, values)
			except VarLibMergeError as e:
				e.stack.append('[%d]' % i)
				raise

	def mergeThings(self, out, lst):
		if not allEqualTo(out, lst, type):
			raise MismatchedTypes(self,
					expected=type(out).__name__,
					got=[type(x).__name__ for x in lst]
			)
		mergerFunc = self.mergersFor(out).get(None, None)
		if mergerFunc is not None:
			mergerFunc(self, out, lst)
		elif hasattr(out, '__dict__'):
			self.mergeObjects(out, lst)
		elif isinstance(out, list):
			self.mergeLists(out, lst)
		else:
			if not allEqualTo(out, lst):
				raise ShouldBeConstant(self, expected=out, got=lst)

	def mergeTables(self, font, master_ttfs, tableTags):
		for tag in tableTags:
			if tag not in font: continue
			try:
				self.ttfs = [m for m in master_ttfs if tag in m]
				self.mergeThings(font[tag], [m[tag] if tag in m else None
							     for m in master_ttfs])
			except VarLibMergeError as e:
				e.stack.append(tag)
				raise

#
# Aligning merger
#
class AligningMerger(Merger):
	pass

@AligningMerger.merger(ot.GDEF, "GlyphClassDef")
def merge(merger, self, lst):
	if self is None:
		if not allNone(lst):
			raise NotANone(merger, expected=None, got=lst)
		return

	lst = [l.classDefs for l in lst]
	self.classDefs = {}
	# We only care about the .classDefs
	self = self.classDefs

	allKeys = set()
	allKeys.update(*[l.keys() for l in lst])
	for k in allKeys:
		allValues = nonNone(l.get(k) for l in lst)
		if not allEqual(allValues):
			raise ShouldBeConstant(merger, expected=allValues[0], got=lst, stack=["." + k])
		if not allValues:
			self[k] = None
		else:
			self[k] = allValues[0]

def _SinglePosUpgradeToFormat2(self):
	if self.Format == 2: return self

	ret = ot.SinglePos()
	ret.Format = 2
	ret.Coverage = self.Coverage
	ret.ValueFormat = self.ValueFormat
	ret.Value = [self.Value for _ in ret.Coverage.glyphs]
	ret.ValueCount = len(ret.Value)

	return ret

def _merge_GlyphOrders(font, lst, values_lst=None, default=None):
	"""Takes font and list of glyph lists (must be sorted by glyph id), and returns
	two things:
	- Combined glyph list,
	- If values_lst is None, return input glyph lists, but padded with None when a glyph
	  was missing in a list.  Otherwise, return values_lst list-of-list, padded with None
	  to match combined glyph lists.
	"""
	if values_lst is None:
		dict_sets = [set(l) for l in lst]
	else:
		dict_sets = [{g:v for g,v in zip(l,vs)} for l,vs in zip(lst,values_lst)]
	combined = set()
	combined.update(*dict_sets)

	sortKey = font.getReverseGlyphMap().__getitem__
	order = sorted(combined, key=sortKey)
	# Make sure all input glyphsets were in proper order
	if not all(sorted(vs, key=sortKey) == vs for vs in lst):
		raise InconsistentGlyphOrder()
	del combined

	paddedValues = None
	if values_lst is None:
		padded = [[glyph if glyph in dict_set else default
			   for glyph in order]
			  for dict_set in dict_sets]
	else:
		assert len(lst) == len(values_lst)
		padded = [[dict_set[glyph] if glyph in dict_set else default
			   for glyph in order]
			  for dict_set in dict_sets]
	return order, padded

def _Lookup_SinglePos_get_effective_value(merger, subtables, glyph):
	for self in subtables:
		if self is None or \
		   type(self) != ot.SinglePos or \
		   self.Coverage is None or \
		   glyph not in self.Coverage.glyphs:
			continue
		if self.Format == 1:
			return self.Value
		elif self.Format == 2:
			return self.Value[self.Coverage.glyphs.index(glyph)]
		else:
			raise UnsupportedFormat(merger, subtable="single positioning lookup")
	return None

def _Lookup_PairPos_get_effective_value_pair(merger, subtables, firstGlyph, secondGlyph):
	for self in subtables:
		if self is None or \
		   type(self) != ot.PairPos or \
		   self.Coverage is None or \
		   firstGlyph not in self.Coverage.glyphs:
			continue
		if self.Format == 1:
			ps = self.PairSet[self.Coverage.glyphs.index(firstGlyph)]
			pvr = ps.PairValueRecord
			for rec in pvr: # TODO Speed up
				if rec.SecondGlyph == secondGlyph:
					return rec
			continue
		elif self.Format == 2:
			klass1 = self.ClassDef1.classDefs.get(firstGlyph, 0)
			klass2 = self.ClassDef2.classDefs.get(secondGlyph, 0)
			return self.Class1Record[klass1].Class2Record[klass2]
		else:
			raise UnsupportedFormat(merger, subtable="pair positioning lookup")
	return None

@AligningMerger.merger(ot.SinglePos)
def merge(merger, self, lst):
	self.ValueFormat = valueFormat = reduce(int.__or__, [l.ValueFormat for l in lst], 0)
	if not (len(lst) == 1 or (valueFormat & ~0xF == 0)):
		raise UnsupportedFormat(merger, subtable="single positioning lookup")

	# If all have same coverage table and all are format 1,
	coverageGlyphs = self.Coverage.glyphs
	if all(v.Format == 1 for v in lst) and all(coverageGlyphs == v.Coverage.glyphs for v in lst):
		self.Value = otBase.ValueRecord(valueFormat)
		merger.mergeThings(self.Value, [v.Value for v in lst])
		self.ValueFormat = self.Value.getFormat()
		return

	# Upgrade everything to Format=2
	self.Format = 2
	lst = [_SinglePosUpgradeToFormat2(v) for v in lst]

	# Align them
	glyphs, padded = _merge_GlyphOrders(merger.font,
					    [v.Coverage.glyphs for v in lst],
					    [v.Value for v in lst])

	self.Coverage.glyphs = glyphs
	self.Value = [otBase.ValueRecord(valueFormat) for _ in glyphs]
	self.ValueCount = len(self.Value)

	for i,values in enumerate(padded):
		for j,glyph in enumerate(glyphs):
			if values[j] is not None: continue
			# Fill in value from other subtables
			# Note!!! This *might* result in behavior change if ValueFormat2-zeroedness
			# is different between used subtable and current subtable!
			# TODO(behdad) Check and warn if that happens?
			v = _Lookup_SinglePos_get_effective_value(merger, merger.lookup_subtables[i], glyph)
			if v is None:
				v = otBase.ValueRecord(valueFormat)
			values[j] = v

	merger.mergeLists(self.Value, padded)

	# Merge everything else; though, there shouldn't be anything else. :)
	merger.mergeObjects(self, lst,
			    exclude=('Format', 'Coverage', 'Value', 'ValueCount'))
	self.ValueFormat = reduce(int.__or__, [v.getEffectiveFormat() for v in self.Value], 0)

@AligningMerger.merger(ot.PairSet)
def merge(merger, self, lst):
	# Align them
	glyphs, padded = _merge_GlyphOrders(merger.font,
				[[v.SecondGlyph for v in vs.PairValueRecord] for vs in lst],
				[vs.PairValueRecord for vs in lst])

	self.PairValueRecord = pvrs = []
	for glyph in glyphs:
		pvr = ot.PairValueRecord()
		pvr.SecondGlyph = glyph
		pvr.Value1 = otBase.ValueRecord(merger.valueFormat1) if merger.valueFormat1 else None
		pvr.Value2 = otBase.ValueRecord(merger.valueFormat2) if merger.valueFormat2 else None
		pvrs.append(pvr)
	self.PairValueCount = len(self.PairValueRecord)

	for i,values in enumerate(padded):
		for j,glyph in enumerate(glyphs):
			# Fill in value from other subtables
			v = ot.PairValueRecord()
			v.SecondGlyph = glyph
			if values[j] is not None:
				vpair = values[j]
			else:
				vpair = _Lookup_PairPos_get_effective_value_pair(
					merger, merger.lookup_subtables[i], self._firstGlyph, glyph
				)
			if vpair is None:
				v1, v2 = None, None
			else:
				v1 = getattr(vpair, "Value1", None)
				v2 = getattr(vpair, "Value2", None)
			v.Value1 = otBase.ValueRecord(merger.valueFormat1, src=v1) if merger.valueFormat1 else None
			v.Value2 = otBase.ValueRecord(merger.valueFormat2, src=v2) if merger.valueFormat2 else None
			values[j] = v
	del self._firstGlyph

	merger.mergeLists(self.PairValueRecord, padded)

def _PairPosFormat1_merge(self, lst, merger):
	assert allEqual([l.ValueFormat2 == 0 for l in lst if l.PairSet]), "Report bug against fonttools."

	# Merge everything else; makes sure Format is the same.
	merger.mergeObjects(self, lst,
			    exclude=('Coverage',
				     'PairSet', 'PairSetCount',
				     'ValueFormat1', 'ValueFormat2'))

	empty = ot.PairSet()
	empty.PairValueRecord = []
	empty.PairValueCount = 0

	# Align them
	glyphs, padded = _merge_GlyphOrders(merger.font,
					    [v.Coverage.glyphs for v in lst],
					    [v.PairSet for v in lst],
					    default=empty)

	self.Coverage.glyphs = glyphs
	self.PairSet = [ot.PairSet() for _ in glyphs]
	self.PairSetCount = len(self.PairSet)
	for glyph, ps in zip(glyphs, self.PairSet):
		ps._firstGlyph = glyph

	merger.mergeLists(self.PairSet, padded)

def _ClassDef_invert(self, allGlyphs=None):

	if isinstance(self, dict):
		classDefs = self
	else:
		classDefs = self.classDefs if self and self.classDefs else {}
	m = max(classDefs.values()) if classDefs else 0

	ret = []
	for _ in range(m + 1):
		ret.append(set())

	for k,v in classDefs.items():
		ret[v].add(k)

	# Class-0 is special.  It's "everything else".
	if allGlyphs is None:
		ret[0] = None
	else:
		# Limit all classes to glyphs in allGlyphs.
		# Collect anything without a non-zero class into class=zero.
		ret[0] = class0 = set(allGlyphs)
		for s in ret[1:]:
			s.intersection_update(class0)
			class0.difference_update(s)

	return ret

def _ClassDef_merge_classify(lst, allGlyphses=None):
	self = ot.ClassDef()
	self.classDefs = classDefs = {}
	allGlyphsesWasNone = allGlyphses is None
	if allGlyphsesWasNone:
		allGlyphses = [None] * len(lst)

	classifier = classifyTools.Classifier()
	for classDef,allGlyphs in zip(lst, allGlyphses):
		sets = _ClassDef_invert(classDef, allGlyphs)
		if allGlyphs is None:
			sets = sets[1:]
		classifier.update(sets)
	classes = classifier.getClasses()

	if allGlyphsesWasNone:
		classes.insert(0, set())

	for i,classSet in enumerate(classes):
		if i == 0:
			continue
		for g in classSet:
			classDefs[g] = i

	return self, classes

def _PairPosFormat2_align_matrices(self, lst, font, transparent=False):

	matrices = [l.Class1Record for l in lst]

	# Align first classes
	self.ClassDef1, classes = _ClassDef_merge_classify([l.ClassDef1 for l in lst], [l.Coverage.glyphs for l in lst])
	self.Class1Count = len(classes)
	new_matrices = []
	for l,matrix in zip(lst, matrices):
		nullRow = None
		coverage = set(l.Coverage.glyphs)
		classDef1 = l.ClassDef1.classDefs
		class1Records = []
		for classSet in classes:
			exemplarGlyph = next(iter(classSet))
			if exemplarGlyph not in coverage:
				# Follow-up to e6125b353e1f54a0280ded5434b8e40d042de69f,
				# Fixes https://github.com/googlei18n/fontmake/issues/470
				# Again, revert 8d441779e5afc664960d848f62c7acdbfc71d7b9
				# when merger becomes selfless.
				nullRow = None
				if nullRow is None:
					nullRow = ot.Class1Record()
					class2records = nullRow.Class2Record = []
					# TODO: When merger becomes selfless, revert e6125b353e1f54a0280ded5434b8e40d042de69f
					for _ in range(l.Class2Count):
						if transparent:
							rec2 = None
						else:
							rec2 = ot.Class2Record()
							rec2.Value1 = otBase.ValueRecord(self.ValueFormat1) if self.ValueFormat1 else None
							rec2.Value2 = otBase.ValueRecord(self.ValueFormat2) if self.ValueFormat2 else None
						class2records.append(rec2)
				rec1 = nullRow
			else:
				klass = classDef1.get(exemplarGlyph, 0)
				rec1 = matrix[klass] # TODO handle out-of-range?
			class1Records.append(rec1)
		new_matrices.append(class1Records)
	matrices = new_matrices
	del new_matrices

	# Align second classes
	self.ClassDef2, classes = _ClassDef_merge_classify([l.ClassDef2 for l in lst])
	self.Class2Count = len(classes)
	new_matrices = []
	for l,matrix in zip(lst, matrices):
		classDef2 = l.ClassDef2.classDefs
		class1Records = []
		for rec1old in matrix:
			oldClass2Records = rec1old.Class2Record
			rec1new = ot.Class1Record()
			class2Records = rec1new.Class2Record = []
			for classSet in classes:
				if not classSet: # class=0
					rec2 = oldClass2Records[0]
				else:
					exemplarGlyph = next(iter(classSet))
					klass = classDef2.get(exemplarGlyph, 0)
					rec2 = oldClass2Records[klass]
				class2Records.append(copy.deepcopy(rec2))
			class1Records.append(rec1new)
		new_matrices.append(class1Records)
	matrices = new_matrices
	del new_matrices

	return matrices

def _PairPosFormat2_merge(self, lst, merger):
	assert allEqual([l.ValueFormat2 == 0 for l in lst if l.Class1Record]), "Report bug against fonttools."

	merger.mergeObjects(self, lst,
			    exclude=('Coverage',
				     'ClassDef1', 'Class1Count',
				     'ClassDef2', 'Class2Count',
				     'Class1Record',
				     'ValueFormat1', 'ValueFormat2'))

	# Align coverages
	glyphs, _ = _merge_GlyphOrders(merger.font,
				       [v.Coverage.glyphs for v in lst])
	self.Coverage.glyphs = glyphs

	# Currently, if the coverage of PairPosFormat2 subtables are different,
	# we do NOT bother walking down the subtable list when filling in new
	# rows for alignment.  As such, this is only correct if current subtable
	# is the last subtable in the lookup.  Ensure that.
	#
	# Note that our canonicalization process merges trailing PairPosFormat2's,
	# so in reality this is rare.
	for l,subtables in zip(lst,merger.lookup_subtables):
		if l.Coverage.glyphs != glyphs:
			assert l == subtables[-1]

	matrices = _PairPosFormat2_align_matrices(self, lst, merger.font)

	self.Class1Record = list(matrices[0]) # TODO move merger to be selfless
	merger.mergeLists(self.Class1Record, matrices)

@AligningMerger.merger(ot.PairPos)
def merge(merger, self, lst):
	merger.valueFormat1 = self.ValueFormat1 = reduce(int.__or__, [l.ValueFormat1 for l in lst], 0)
	merger.valueFormat2 = self.ValueFormat2 = reduce(int.__or__, [l.ValueFormat2 for l in lst], 0)

	if self.Format == 1:
		_PairPosFormat1_merge(self, lst, merger)
	elif self.Format == 2:
		_PairPosFormat2_merge(self, lst, merger)
	else:
		raise UnsupportedFormat(merger, subtable="pair positioning lookup")

	del merger.valueFormat1, merger.valueFormat2

	# Now examine the list of value records, and update to the union of format values,
	# as merge might have created new values.
	vf1 = 0
	vf2 = 0
	if self.Format == 1:
		for pairSet in self.PairSet:
			for pairValueRecord in pairSet.PairValueRecord:
				pv1 = getattr(pairValueRecord, "Value1", None)
				if pv1 is not None:
					vf1 |= pv1.getFormat()
				pv2 = getattr(pairValueRecord, "Value2", None)
				if pv2 is not None:
					vf2 |= pv2.getFormat()
	elif self.Format == 2:
		for class1Record in self.Class1Record:
			for class2Record in class1Record.Class2Record:
				pv1 = getattr(class2Record, "Value1", None)
				if pv1 is not None:
					vf1 |= pv1.getFormat()
				pv2 = getattr(class2Record, "Value2", None)
				if pv2 is not None:
					vf2 |= pv2.getFormat()
	self.ValueFormat1 = vf1
	self.ValueFormat2 = vf2

def _MarkBasePosFormat1_merge(self, lst, merger, Mark='Mark', Base='Base'):
	self.ClassCount = max(l.ClassCount for l in lst)

	MarkCoverageGlyphs, MarkRecords = \
		_merge_GlyphOrders(merger.font,
				   [getattr(l, Mark+'Coverage').glyphs for l in lst],
				   [getattr(l, Mark+'Array').MarkRecord for l in lst])
	getattr(self, Mark+'Coverage').glyphs = MarkCoverageGlyphs

	BaseCoverageGlyphs, BaseRecords = \
		_merge_GlyphOrders(merger.font,
				   [getattr(l, Base+'Coverage').glyphs for l in lst],
				   [getattr(getattr(l, Base+'Array'), Base+'Record') for l in lst])
	getattr(self, Base+'Coverage').glyphs = BaseCoverageGlyphs

	# MarkArray
	records = []
	for g,glyphRecords in zip(MarkCoverageGlyphs, zip(*MarkRecords)):
		allClasses = [r.Class for r in glyphRecords if r is not None]

		# TODO Right now we require that all marks have same class in
		# all masters that cover them.  This is not required.
		#
		# We can relax that by just requiring that all marks that have
		# the same class in a master, have the same class in every other
		# master.  Indeed, if, say, a sparse master only covers one mark,
		# that mark probably will get class 0, which would possibly be
		# different from its class in other masters.
		#
		# We can even go further and reclassify marks to support any
		# input.  But, since, it's unlikely that two marks being both,
		# say, "top" in one master, and one being "top" and other being
		# "top-right" in another master, we shouldn't do that, as any
		# failures in that case will probably signify mistakes in the
		# input masters.

		if not allEqual(allClasses):
			raise ShouldBeConstant(merger, expected=allClasses[0], got=allClasses)
		else:
			rec = ot.MarkRecord()
			rec.Class = allClasses[0]
			allAnchors = [None if r is None else r.MarkAnchor for r in glyphRecords]
			if allNone(allAnchors):
				anchor = None
			else:
				anchor = ot.Anchor()
				anchor.Format = 1
				merger.mergeThings(anchor, allAnchors)
			rec.MarkAnchor = anchor
		records.append(rec)
	array = ot.MarkArray()
	array.MarkRecord = records
	array.MarkCount = len(records)
	setattr(self, Mark+"Array", array)

	# BaseArray
	records = []
	for g,glyphRecords in zip(BaseCoverageGlyphs, zip(*BaseRecords)):
		if allNone(glyphRecords):
			rec = None
		else:
			rec = getattr(ot, Base+'Record')()
			anchors = []
			setattr(rec, Base+'Anchor', anchors)
			glyphAnchors = [[] if r is None else getattr(r, Base+'Anchor')
					for r in glyphRecords]
			for l in glyphAnchors:
				l.extend([None] * (self.ClassCount - len(l)))
			for allAnchors in zip(*glyphAnchors):
				if allNone(allAnchors):
					anchor = None
				else:
					anchor = ot.Anchor()
					anchor.Format = 1
					merger.mergeThings(anchor, allAnchors)
				anchors.append(anchor)
		records.append(rec)
	array = getattr(ot, Base+'Array')()
	setattr(array, Base+'Record', records)
	setattr(array, Base+'Count', len(records))
	setattr(self, Base+'Array', array)

@AligningMerger.merger(ot.MarkBasePos)
def merge(merger, self, lst):
	if not allEqualTo(self.Format, (l.Format for l in lst)):
		raise InconsistentFormats(
			merger,
			subtable="mark-to-base positioning lookup",
			expected=self.Format,
			got=[l.Format for l in lst]
		)
	if self.Format == 1:
		_MarkBasePosFormat1_merge(self, lst, merger)
	else:
		raise UnsupportedFormat(merger, subtable="mark-to-base positioning lookup")

@AligningMerger.merger(ot.MarkMarkPos)
def merge(merger, self, lst):
	if not allEqualTo(self.Format, (l.Format for l in lst)):
		raise InconsistentFormats(
			merger,
			subtable="mark-to-mark positioning lookup",
			expected=self.Format,
			got=[l.Format for l in lst]
		)
	if self.Format == 1:
		_MarkBasePosFormat1_merge(self, lst, merger, 'Mark1', 'Mark2')
	else:
		raise UnsupportedFormat(merger, subtable="mark-to-mark positioning lookup")

def _PairSet_flatten(lst, font):
	self = ot.PairSet()
	self.Coverage = ot.Coverage()

	# Align them
	glyphs, padded = _merge_GlyphOrders(font,
				[[v.SecondGlyph for v in vs.PairValueRecord] for vs in lst],
				[vs.PairValueRecord for vs in lst])

	self.Coverage.glyphs = glyphs
	self.PairValueRecord = pvrs = []
	for values in zip(*padded):
		for v in values:
			if v is not None:
				pvrs.append(v)
				break
		else:
			assert False
	self.PairValueCount = len(self.PairValueRecord)

	return self

def _Lookup_PairPosFormat1_subtables_flatten(lst, font):
	assert allEqual([l.ValueFormat2 == 0 for l in lst if l.PairSet]), "Report bug against fonttools."

	self = ot.PairPos()
	self.Format = 1
	self.Coverage = ot.Coverage()
	self.ValueFormat1 = reduce(int.__or__, [l.ValueFormat1 for l in lst], 0)
	self.ValueFormat2 = reduce(int.__or__, [l.ValueFormat2 for l in lst], 0)

	# Align them
	glyphs, padded = _merge_GlyphOrders(font,
					    [v.Coverage.glyphs for v in lst],
					    [v.PairSet for v in lst])

	self.Coverage.glyphs = glyphs
	self.PairSet = [_PairSet_flatten([v for v in values if v is not None], font)
		        for values in zip(*padded)]
	self.PairSetCount = len(self.PairSet)
	return self

def _Lookup_PairPosFormat2_subtables_flatten(lst, font):
	assert allEqual([l.ValueFormat2 == 0 for l in lst if l.Class1Record]), "Report bug against fonttools."

	self = ot.PairPos()
	self.Format = 2
	self.Coverage = ot.Coverage()
	self.ValueFormat1 = reduce(int.__or__, [l.ValueFormat1 for l in lst], 0)
	self.ValueFormat2 = reduce(int.__or__, [l.ValueFormat2 for l in lst], 0)

	# Align them
	glyphs, _ = _merge_GlyphOrders(font,
				       [v.Coverage.glyphs for v in lst])
	self.Coverage.glyphs = glyphs

	matrices = _PairPosFormat2_align_matrices(self, lst, font, transparent=True)

	matrix = self.Class1Record = []
	for rows in zip(*matrices):
		row = ot.Class1Record()
		matrix.append(row)
		row.Class2Record = []
		row = row.Class2Record
		for cols in zip(*list(r.Class2Record for r in rows)):
			col = next(iter(c for c in cols if c is not None))
			row.append(col)

	return self

def _Lookup_PairPos_subtables_canonicalize(lst, font):
	"""Merge multiple Format1 subtables at the beginning of lst,
	and merge multiple consecutive Format2 subtables that have the same
	Class2 (ie. were split because of offset overflows).  Returns new list."""
	lst = list(lst)

	l = len(lst)
	i = 0
	while i < l and lst[i].Format == 1:
		i += 1
	lst[:i] = [_Lookup_PairPosFormat1_subtables_flatten(lst[:i], font)]

	l = len(lst)
	i = l
	while i > 0 and lst[i - 1].Format == 2:
		i -= 1
	lst[i:] = [_Lookup_PairPosFormat2_subtables_flatten(lst[i:], font)]

	return lst

def _Lookup_SinglePos_subtables_flatten(lst, font, min_inclusive_rec_format):
	glyphs, _ = _merge_GlyphOrders(font,
		[v.Coverage.glyphs for v in lst], None)
	num_glyphs = len(glyphs)
	new = ot.SinglePos()
	new.Format = 2
	new.ValueFormat = min_inclusive_rec_format
	new.Coverage = ot.Coverage()
	new.Coverage.glyphs = glyphs
	new.ValueCount = num_glyphs
	new.Value = [None] * num_glyphs
	for singlePos in lst:
		if singlePos.Format == 1:
			val_rec = singlePos.Value
			for gname in singlePos.Coverage.glyphs:
				i = glyphs.index(gname)
				new.Value[i] = copy.deepcopy(val_rec)
		elif singlePos.Format == 2:
			for j, gname in enumerate(singlePos.Coverage.glyphs):
				val_rec = singlePos.Value[j]
				i = glyphs.index(gname)
				new.Value[i] = copy.deepcopy(val_rec)
	return [new]

@AligningMerger.merger(ot.Lookup)
def merge(merger, self, lst):
	subtables = merger.lookup_subtables = [l.SubTable for l in lst]

	# Remove Extension subtables
	for l,sts in list(zip(lst,subtables))+[(self,self.SubTable)]:
		if not sts:
			continue
		if sts[0].__class__.__name__.startswith('Extension'):
			if not allEqual([st.__class__ for st in sts]):
				raise InconsistentExtensions(
					merger,
					expected="Extension",
					got=[st.__class__.__name__ for st in sts]
				)
			if not allEqual([st.ExtensionLookupType for st in sts]):
				raise InconsistentExtensions(merger)
			l.LookupType = sts[0].ExtensionLookupType
			new_sts = [st.ExtSubTable for st in sts]
			del sts[:]
			sts.extend(new_sts)

	isPairPos = self.SubTable and isinstance(self.SubTable[0], ot.PairPos)

	if isPairPos:
		# AFDKO and feaLib sometimes generate two Format1 subtables instead of one.
		# Merge those before continuing.
		# https://github.com/fonttools/fonttools/issues/719
		self.SubTable = _Lookup_PairPos_subtables_canonicalize(self.SubTable, merger.font)
		subtables = merger.lookup_subtables = [_Lookup_PairPos_subtables_canonicalize(st, merger.font) for st in subtables]
	else:
		isSinglePos = self.SubTable and isinstance(self.SubTable[0], ot.SinglePos)
		if isSinglePos:
			numSubtables = [len(st) for st in subtables]
			if not all([nums == numSubtables[0] for nums in numSubtables]):
				# Flatten list of SinglePos subtables to single Format 2 subtable,
				# with all value records set to the rec format type.
				# We use buildSinglePos() to optimize the lookup after merging.
				valueFormatList = [t.ValueFormat for st in subtables for t in st]
				# Find the minimum value record that can accomodate all the singlePos subtables.
				mirf = reduce(ior, valueFormatList)
				self.SubTable = _Lookup_SinglePos_subtables_flatten(self.SubTable, merger.font, mirf)
				subtables = merger.lookup_subtables = [
					_Lookup_SinglePos_subtables_flatten(st, merger.font, mirf) for st in subtables]
				flattened = True
			else:
				flattened = False

	merger.mergeLists(self.SubTable, subtables)
	self.SubTableCount = len(self.SubTable)

	if isPairPos:
		# If format-1 subtable created during canonicalization is empty, remove it.
		assert len(self.SubTable) >= 1 and self.SubTable[0].Format == 1
		if not self.SubTable[0].Coverage.glyphs:
			self.SubTable.pop(0)
			self.SubTableCount -= 1

		# If format-2 subtable created during canonicalization is empty, remove it.
		assert len(self.SubTable) >= 1 and self.SubTable[-1].Format == 2
		if not self.SubTable[-1].Coverage.glyphs:
			self.SubTable.pop(-1)
			self.SubTableCount -= 1

	elif isSinglePos and flattened:
		singlePosTable = self.SubTable[0]
		glyphs = singlePosTable.Coverage.glyphs
		# We know that singlePosTable is Format 2, as this is set
		# in _Lookup_SinglePos_subtables_flatten.
		singlePosMapping = {
			gname: valRecord
			for gname, valRecord in zip(glyphs, singlePosTable.Value)
		}
		self.SubTable = buildSinglePos(singlePosMapping, merger.font.getReverseGlyphMap())
	merger.mergeObjects(self, lst, exclude=['SubTable', 'SubTableCount'])

	del merger.lookup_subtables


#
# InstancerMerger
#

class InstancerMerger(AligningMerger):
	"""A merger that takes multiple master fonts, and instantiates
	an instance."""

	def __init__(self, font, model, location):
		Merger.__init__(self, font)
		self.model = model
		self.location = location
		self.scalars = model.getScalars(location)

@InstancerMerger.merger(ot.CaretValue)
def merge(merger, self, lst):
	assert self.Format == 1
	Coords = [a.Coordinate for a in lst]
	model = merger.model
	scalars = merger.scalars
	self.Coordinate = otRound(model.interpolateFromMastersAndScalars(Coords, scalars))

@InstancerMerger.merger(ot.Anchor)
def merge(merger, self, lst):
	assert self.Format == 1
	XCoords = [a.XCoordinate for a in lst]
	YCoords = [a.YCoordinate for a in lst]
	model = merger.model
	scalars = merger.scalars
	self.XCoordinate = otRound(model.interpolateFromMastersAndScalars(XCoords, scalars))
	self.YCoordinate = otRound(model.interpolateFromMastersAndScalars(YCoords, scalars))

@InstancerMerger.merger(otBase.ValueRecord)
def merge(merger, self, lst):
	model = merger.model
	scalars = merger.scalars
	# TODO Handle differing valueformats
	for name, tableName in [('XAdvance','XAdvDevice'),
				('YAdvance','YAdvDevice'),
				('XPlacement','XPlaDevice'),
				('YPlacement','YPlaDevice')]:

		assert not hasattr(self, tableName)

		if hasattr(self, name):
			values = [getattr(a, name, 0) for a in lst]
			value = otRound(model.interpolateFromMastersAndScalars(values, scalars))
			setattr(self, name, value)


#
# MutatorMerger
#

class MutatorMerger(AligningMerger):
	"""A merger that takes a variable font, and instantiates
	an instance.  While there's no "merging" to be done per se,
	the operation can benefit from many operations that the
	aligning merger does."""

	def __init__(self, font, instancer, deleteVariations=True):
		Merger.__init__(self, font)
		self.instancer = instancer
		self.deleteVariations = deleteVariations

@MutatorMerger.merger(ot.CaretValue)
def merge(merger, self, lst):

	# Hack till we become selfless.
	self.__dict__ = lst[0].__dict__.copy()

	if self.Format != 3:
		return

	instancer = merger.instancer
	dev = self.DeviceTable
	if merger.deleteVariations:
		del self.DeviceTable
	if dev:
		assert dev.DeltaFormat == 0x8000
		varidx = (dev.StartSize << 16) + dev.EndSize
		delta = otRound(instancer[varidx])
		self.Coordinate += delta

	if merger.deleteVariations:
		self.Format = 1

@MutatorMerger.merger(ot.Anchor)
def merge(merger, self, lst):

	# Hack till we become selfless.
	self.__dict__ = lst[0].__dict__.copy()

	if self.Format != 3:
		return

	instancer = merger.instancer
	for v in "XY":
		tableName = v+'DeviceTable'
		if not hasattr(self, tableName):
			continue
		dev = getattr(self, tableName)
		if merger.deleteVariations:
			delattr(self, tableName)
		if dev is None:
			continue

		assert dev.DeltaFormat == 0x8000
		varidx = (dev.StartSize << 16) + dev.EndSize
		delta = otRound(instancer[varidx])

		attr = v+'Coordinate'
		setattr(self, attr, getattr(self, attr) + delta)

	if merger.deleteVariations:
		self.Format = 1

@MutatorMerger.merger(otBase.ValueRecord)
def merge(merger, self, lst):

	# Hack till we become selfless.
	self.__dict__ = lst[0].__dict__.copy()

	instancer = merger.instancer
	for name, tableName in [('XAdvance','XAdvDevice'),
				('YAdvance','YAdvDevice'),
				('XPlacement','XPlaDevice'),
				('YPlacement','YPlaDevice')]:

		if not hasattr(self, tableName):
			continue
		dev = getattr(self, tableName)
		if merger.deleteVariations:
			delattr(self, tableName)
		if dev is None:
			continue

		assert dev.DeltaFormat == 0x8000
		varidx = (dev.StartSize << 16) + dev.EndSize
		delta = otRound(instancer[varidx])

		setattr(self, name, getattr(self, name) + delta)


#
# VariationMerger
#

class VariationMerger(AligningMerger):
	"""A merger that takes multiple master fonts, and builds a
	variable font."""

	def __init__(self, model, axisTags, font):
		Merger.__init__(self, font)
		self.store_builder = varStore.OnlineVarStoreBuilder(axisTags)
		self.setModel(model)

	def setModel(self, model):
		self.model = model
		self.store_builder.setModel(model)

	def mergeThings(self, out, lst):
		masterModel = None
		if None in lst:
			if allNone(lst):
				if out is not None:
					raise FoundANone(self, got=lst)
				return
			masterModel = self.model
			model, lst = masterModel.getSubModel(lst)
			self.setModel(model)

		super(VariationMerger, self).mergeThings(out, lst)

		if masterModel:
			self.setModel(masterModel)


def buildVarDevTable(store_builder, master_values):
	if allEqual(master_values):
		return master_values[0], None
	base, varIdx = store_builder.storeMasters(master_values)
	return base, builder.buildVarDevTable(varIdx)

@VariationMerger.merger(ot.BaseCoord)
def merge(merger, self, lst):
	if self.Format != 1:
		raise UnsupportedFormat(merger, subtable="a baseline coordinate")
	self.Coordinate, DeviceTable = buildVarDevTable(merger.store_builder, [a.Coordinate for a in lst])
	if DeviceTable:
		self.Format = 3
		self.DeviceTable = DeviceTable

@VariationMerger.merger(ot.CaretValue)
def merge(merger, self, lst):
	if self.Format != 1:
		raise UnsupportedFormat(merger, subtable="a caret")
	self.Coordinate, DeviceTable = buildVarDevTable(merger.store_builder, [a.Coordinate for a in lst])
	if DeviceTable:
		self.Format = 3
		self.DeviceTable = DeviceTable

@VariationMerger.merger(ot.Anchor)
def merge(merger, self, lst):
	if self.Format != 1:
		raise UnsupportedFormat(merger, subtable="an anchor")
	self.XCoordinate, XDeviceTable = buildVarDevTable(merger.store_builder, [a.XCoordinate for a in lst])
	self.YCoordinate, YDeviceTable = buildVarDevTable(merger.store_builder, [a.YCoordinate for a in lst])
	if XDeviceTable or YDeviceTable:
		self.Format = 3
		self.XDeviceTable = XDeviceTable
		self.YDeviceTable = YDeviceTable

@VariationMerger.merger(otBase.ValueRecord)
def merge(merger, self, lst):
	for name, tableName in [('XAdvance','XAdvDevice'),
				('YAdvance','YAdvDevice'),
				('XPlacement','XPlaDevice'),
				('YPlacement','YPlaDevice')]:

		if hasattr(self, name):
			value, deviceTable = buildVarDevTable(merger.store_builder,
							      [getattr(a, name, 0) for a in lst])
			setattr(self, name, value)
			if deviceTable:
				setattr(self, tableName, deviceTable)
